from unittest import TestCase

import pytest

from miio import AirFreshA1, AirFreshT2017
from miio.airfresh_t2017 import (
    MODEL_AIRFRESH_A1,
    MODEL_AIRFRESH_T2017,
    AirFreshException,
    AirFreshStatus,
    DisplayOrientation,
    OperationMode,
    PtcLevel,
)

from .dummies import DummyDevice


class DummyAirFreshA1(DummyDevice, AirFreshA1):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_AIRFRESH_A1
        self.state = {
            "power": True,
            "mode": "auto",
            "pm25": 2,
            "co2": 554,
            "temperature_outside": 12,
            "favourite_speed": 150,
            "control_speed": 45,
            "filter_rate": 45,
            "filter_day": 81,
            "ptc_on": False,
            "ptc_status": False,
            "child_lock": False,
            "sound": True,
            "display": False,
        }
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
            "set_mode": lambda x: self._set_state("mode", x),
            "set_sound": lambda x: self._set_state("sound", x),
            "set_child_lock": lambda x: self._set_state("child_lock", x),
            "set_display": lambda x: self._set_state("display", x),
            "set_ptc_on": lambda x: self._set_state("ptc_on", x),
            "set_favourite_speed": lambda x: self._set_state("favourite_speed", x),
            "set_filter_rate": lambda x: self._set_filter_rate(x),
        }
        super().__init__(args, kwargs)

    def _set_filter_rate(self, value: str):
        if value[0] == 100:
            self._set_state("filter_rate", [100])
            self._set_state("filter_day", [180])


@pytest.fixture(scope="class")
def airfresha1(request):
    request.cls.device = DummyAirFreshA1()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("airfresha1")
class TestAirFreshA1(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(AirFreshStatus(self.device.start_state))

        assert self.is_on() is True
        assert (
            self.state().temperature == self.device.start_state["temperature_outside"]
        )
        assert self.state().co2 == self.device.start_state["co2"]
        assert self.state().pm25 == self.device.start_state["pm25"]
        assert self.state().mode == OperationMode(self.device.start_state["mode"])
        assert self.state().buzzer == self.device.start_state["sound"]
        assert self.state().child_lock == self.device.start_state["child_lock"]

    def test_set_mode(self):
        def mode():
            return self.device.status().mode

        self.device.set_mode(OperationMode.Off)
        assert mode() == OperationMode.Off

        self.device.set_mode(OperationMode.Auto)
        assert mode() == OperationMode.Auto

        self.device.set_mode(OperationMode.Sleep)
        assert mode() == OperationMode.Sleep

        self.device.set_mode(OperationMode.Favorite)
        assert mode() == OperationMode.Favorite

    def test_set_display(self):
        def display():
            return self.device.status().display

        self.device.set_display(True)
        assert display() is True

        self.device.set_display(False)
        assert display() is False

    def test_set_buzzer(self):
        def buzzer():
            return self.device.status().buzzer

        self.device.set_buzzer(True)
        assert buzzer() is True

        self.device.set_buzzer(False)
        assert buzzer() is False

    def test_set_child_lock(self):
        def child_lock():
            return self.device.status().child_lock

        self.device.set_child_lock(True)
        assert child_lock() is True

        self.device.set_child_lock(False)
        assert child_lock() is False

    def test_reset_dust_filter(self):
        def dust_filter_life_remaining():
            return self.device.status().dust_filter_life_remaining

        def dust_filter_life_remaining_days():
            return self.device.status().dust_filter_life_remaining_days

        self.device._reset_state()
        assert dust_filter_life_remaining() != 100
        assert dust_filter_life_remaining_days() != 180
        self.device.reset_dust_filter()
        assert dust_filter_life_remaining() == 100
        assert dust_filter_life_remaining_days() == 180

    def test_set_favorite_speed(self):
        def favorite_speed():
            return self.device.status().favorite_speed

        self.device.set_favorite_speed(0)
        assert favorite_speed() == 0
        self.device.set_favorite_speed(150)
        assert favorite_speed() == 150

        with pytest.raises(AirFreshException):
            self.device.set_favorite_speed(-1)

        with pytest.raises(AirFreshException):
            self.device.set_favorite_speed(151)

    def test_set_ptc(self):
        def ptc():
            return self.device.status().ptc

        self.device.set_ptc(True)
        assert ptc() is True

        self.device.set_ptc(False)
        assert ptc() is False


class DummyAirFreshT2017(DummyDevice, AirFreshT2017):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_AIRFRESH_T2017
        self.state = {
            "power": True,
            "mode": "favourite",
            "pm25": 1,
            "co2": 550,
            "temperature_outside": 24,
            "favourite_speed": 241,
            "control_speed": 241,
            "filter_intermediate": 99,
            "filter_inter_day": 89,
            "filter_efficient": 99,
            "filter_effi_day": 179,
            "ptc_on": False,
            "ptc_level": "low",
            "ptc_status": False,
            "child_lock": False,
            "sound": True,
            "display": False,
            "screen_direction": "forward",
        }
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
            "set_mode": lambda x: self._set_state("mode", x),
            "set_sound": lambda x: self._set_state("sound", x),
            "set_child_lock": lambda x: self._set_state("child_lock", x),
            "set_display": lambda x: self._set_state("display", x),
            "set_screen_direction": lambda x: self._set_state("screen_direction", x),
            "set_ptc_level": lambda x: self._set_state("ptc_level", x),
            "set_ptc_on": lambda x: self._set_state("ptc_on", x),
            "set_favourite_speed": lambda x: self._set_state("favourite_speed", x),
            "set_filter_reset": lambda x: self._set_filter_reset(x),
        }
        super().__init__(args, kwargs)

    def _set_filter_reset(self, value: str):
        if value[0] == "efficient":
            self._set_state("filter_efficient", [100])
            self._set_state("filter_effi_day", [180])

        if value[0] == "intermediate":
            self._set_state("filter_intermediate", [100])
            self._set_state("filter_inter_day", [90])


@pytest.fixture(scope="class")
def airfresht2017(request):
    request.cls.device = DummyAirFreshT2017()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("airfresht2017")
class TestAirFreshT2017(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(AirFreshStatus(self.device.start_state))

        assert self.is_on() is True
        assert (
            self.state().temperature == self.device.start_state["temperature_outside"]
        )
        assert self.state().co2 == self.device.start_state["co2"]
        assert self.state().pm25 == self.device.start_state["pm25"]
        assert self.state().mode == OperationMode(self.device.start_state["mode"])
        assert self.state().buzzer == self.device.start_state["sound"]
        assert self.state().child_lock == self.device.start_state["child_lock"]

    def test_set_mode(self):
        def mode():
            return self.device.status().mode

        self.device.set_mode(OperationMode.Off)
        assert mode() == OperationMode.Off

        self.device.set_mode(OperationMode.Auto)
        assert mode() == OperationMode.Auto

        self.device.set_mode(OperationMode.Sleep)
        assert mode() == OperationMode.Sleep

        self.device.set_mode(OperationMode.Favorite)
        assert mode() == OperationMode.Favorite

    def test_set_display(self):
        def display():
            return self.device.status().display

        self.device.set_display(True)
        assert display() is True

        self.device.set_display(False)
        assert display() is False

    def test_set_buzzer(self):
        def buzzer():
            return self.device.status().buzzer

        self.device.set_buzzer(True)
        assert buzzer() is True

        self.device.set_buzzer(False)
        assert buzzer() is False

    def test_set_child_lock(self):
        def child_lock():
            return self.device.status().child_lock

        self.device.set_child_lock(True)
        assert child_lock() is True

        self.device.set_child_lock(False)
        assert child_lock() is False

    def test_reset_dust_filter(self):
        def dust_filter_life_remaining():
            return self.device.status().dust_filter_life_remaining

        def dust_filter_life_remaining_days():
            return self.device.status().dust_filter_life_remaining_days

        self.device._reset_state()
        assert dust_filter_life_remaining() != 100
        assert dust_filter_life_remaining_days() != 90
        self.device.reset_dust_filter()
        assert dust_filter_life_remaining() == 100
        assert dust_filter_life_remaining_days() == 90

    def test_reset_upper_filter(self):
        def upper_filter_life_remaining():
            return self.device.status().upper_filter_life_remaining

        def upper_filter_life_remaining_days():
            return self.device.status().upper_filter_life_remaining_days

        self.device._reset_state()
        assert upper_filter_life_remaining() != 100
        assert upper_filter_life_remaining_days() != 180
        self.device.reset_upper_filter()
        assert upper_filter_life_remaining() == 100
        assert upper_filter_life_remaining_days() == 180

    def test_set_favorite_speed(self):
        def favorite_speed():
            return self.device.status().favorite_speed

        self.device.set_favorite_speed(60)
        assert favorite_speed() == 60
        self.device.set_favorite_speed(120)
        assert favorite_speed() == 120
        self.device.set_favorite_speed(240)
        assert favorite_speed() == 240
        self.device.set_favorite_speed(300)
        assert favorite_speed() == 300

        with pytest.raises(AirFreshException):
            self.device.set_favorite_speed(-1)

        with pytest.raises(AirFreshException):
            self.device.set_favorite_speed(59)

        with pytest.raises(AirFreshException):
            self.device.set_favorite_speed(301)

    def test_set_ptc(self):
        def ptc():
            return self.device.status().ptc

        self.device.set_ptc(True)
        assert ptc() is True

        self.device.set_ptc(False)
        assert ptc() is False

    def test_set_ptc_level(self):
        def ptc_level():
            return self.device.status().ptc_level

        self.device.set_ptc_level(PtcLevel.Low)
        assert ptc_level() == PtcLevel.Low
        self.device.set_ptc_level(PtcLevel.Medium)
        assert ptc_level() == PtcLevel.Medium
        self.device.set_ptc_level(PtcLevel.High)
        assert ptc_level() == PtcLevel.High

    def test_set_display_orientation(self):
        def display_orientation():
            return self.device.status().display_orientation

        self.device.set_display_orientation(DisplayOrientation.Portrait)
        assert display_orientation() == DisplayOrientation.Portrait
        self.device.set_display_orientation(DisplayOrientation.LandscapeLeft)
        assert display_orientation() == DisplayOrientation.LandscapeLeft
        self.device.set_display_orientation(DisplayOrientation.LandscapeRight)
        assert display_orientation() == DisplayOrientation.LandscapeRight



from unittest import TestCase

import pytest

from miio import WifiRepeater
from miio.tests.dummies import DummyDevice
from miio.wifirepeater import WifiRepeaterConfiguration, WifiRepeaterStatus


class DummyWifiRepeater(DummyDevice, WifiRepeater):
    def __init__(self, *args, **kwargs):
        self.state = {
            "sta": {"count": 2, "access_policy": 0},
            "mat": [
                {
                    "mac": "aa:aa:aa:aa:aa:aa",
                    "ip": "192.168.1.133",
                    "last_time": 54371873,
                },
                {
                    "mac": "bb:bb:bb:bb:bb:bb",
                    "ip": "192.168.1.156",
                    "last_time": 54371496,
                },
            ],
            "access_list": {"mac": ""},
        }
        self.config = {"ssid": "SSID", "pwd": "PWD", "hidden": 0}
        self.device_info = {
            "life": 543452,
            "cfg_time": 543452,
            "token": "ffffffffffffffffffffffffffffffff",
            "fw_ver": "2.2.14",
            "hw_ver": "R02",
            "uid": 1583412143,
            "api_level": 2,
            "mcu_fw_ver": "1000",
            "wifi_fw_ver": "1.0.0",
            "mac": "FF:FF:FF:FF:FF:FF",
            "model": "xiaomi.repeater.v2",
            "ap": {
                "rssi": -63,
                "ssid": "SSID",
                "bssid": "EE:EE:EE:EE:EE:EE",
                "rx": 136695922,
                "tx": 1779521233,
            },
            "sta": {
                "count": 2,
                "ssid": "REPEATER-SSID",
                "hidden": 0,
                "assoclist": "cc:cc:cc:cc:cc:cc;bb:bb:bb:bb:bb:bb;",
            },
            "netif": {
                "localIp": "192.168.1.170",
                "mask": "255.255.255.0",
                "gw": "192.168.1.1",
            },
            "desc": {
                "wifi_explorer": 1,
                "sn": "14923 / 20191356",
                "color": 101,
                "channel": "release",
            },
        }

        self.return_values = {
            "miIO.get_repeater_sta_info": self._get_state,
            "miIO.get_repeater_ap_info": self._get_configuration,
            "miIO.switch_wifi_explorer": self._set_wifi_explorer,
            "miIO.switch_wifi_ssid": self._set_configuration,
            "miIO.info": self._get_info,
        }
        self.start_state = self.state.copy()
        self.start_config = self.config.copy()
        self.start_device_info = self.device_info.copy()
        super().__init__(args, kwargs)

    def _reset_state(self):
        """Revert back to the original state."""
        self.state = self.start_state.copy()
        self.config = self.start_config.copy()
        self.device_info = self.start_device_info.copy()

    def _get_state(self, param):
        return self.state

    def _get_configuration(self, param):
        return self.config

    def _get_info(self, param):
        return self.device_info

    def _set_wifi_explorer(self, data):
        self.device_info["desc"]["wifi_explorer"] = data[0]["wifi_explorer"]

    def _set_configuration(self, data):
        self.config = {
            "ssid": data[0]["ssid"],
            "pwd": data[0]["pwd"],
            "hidden": data[0]["hidden"],
        }

        self.device_info["desc"]["wifi_explorer"] = data[0]["wifi_explorer"]
        return True


@pytest.fixture(scope="class")
def wifirepeater(request):
    request.cls.device = DummyWifiRepeater()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("wifirepeater")
class TestWifiRepeater(TestCase):
    def state(self):
        return self.device.status()

    def configuration(self):
        return self.device.configuration()

    def info(self):
        return self.device.info()

    def wifi_roaming(self):
        return self.device.wifi_roaming()

    def rssi_accesspoint(self):
        return self.device.rssi_accesspoint()

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(WifiRepeaterStatus(self.device.start_state))

        assert (
            self.state().access_policy
            == self.device.start_state["sta"]["access_policy"]
        )
        assert self.state().associated_stations == self.device.start_state["mat"]

    def test_set_wifi_roaming(self):
        self.device.set_wifi_roaming(True)
        assert self.wifi_roaming() is True

        self.device.set_wifi_roaming(False)
        assert self.wifi_roaming() is False

    def test_configuration(self):
        self.device._reset_state()

        assert repr(self.configuration()) == repr(
            WifiRepeaterConfiguration(self.device.start_config)
        )

        assert self.configuration().ssid == self.device.start_config["ssid"]
        assert self.configuration().password == self.device.start_config["pwd"]
        assert self.configuration().ssid_hidden is (
            self.device.start_config["hidden"] == 1
        )

    def test_set_configuration(self):
        def configuration():
            return self.device.configuration()

        dummy_configuration = {"ssid": "SSID2", "password": "PASSWORD2", "hidden": True}

        self.device.set_configuration(
            dummy_configuration["ssid"],
            dummy_configuration["password"],
            dummy_configuration["hidden"],
        )
        assert configuration().ssid == dummy_configuration["ssid"]
        assert configuration().password == dummy_configuration["password"]
        assert configuration().ssid_hidden is dummy_configuration["hidden"]

    def test_rssi_accesspoint(self):
        self.device._reset_state()

        assert self.rssi_accesspoint() is self.device.start_device_info["ap"]["rssi"]



import enum
import logging
from typing import Any, Dict

import click

from .click_common import EnumType, command, format_output
from .miot_device import DeviceStatus, MiotDevice

_LOGGER = logging.getLogger(__name__)
_MAPPING = {
    # # Source http://miot-spec.org/miot-spec-v2/instance?type=urn:miot-spec-v2:device:curtain:0000A00C:lumi-hagl05:1
    # Curtain
    "motor_control": {"siid": 2, "piid": 2},  # 0 - Pause, 1 - Open, 2 - Close, 3 - auto
    "current_position": {"siid": 2, "piid": 3},  # Range: [0, 100, 1]
    "status": {"siid": 2, "piid": 6},  # 0 - Stopped, 1 - Opening, 2 - Closing
    "target_position": {"siid": 2, "piid": 7},  # Range: [0, 100, 1]
    # curtain_cfg
    "is_manual_enabled": {"siid": 4, "piid": 1},  #
    "polarity": {"siid": 4, "piid": 2},
    "is_position_limited": {"siid": 4, "piid": 3},
    "night_tip_light": {"siid": 4, "piid": 4},
    "run_time": {"siid": 4, "piid": 5},  # Range: [0, 255, 1]
    # motor_controller
    "adjust_value": {"siid": 5, "piid": 1},  # Range: [-100, 100, 1]
}

# Model: ZNCLDJ21LM (also known as "Xiaomiyoupin Curtain Controller (Wi-Fi)"
MODEL_CURTAIN_HAGL05 = "lumi.curtain.hagl05"


class MotorControl(enum.Enum):
    Pause = 0
    Open = 1
    Close = 2
    Auto = 3


class Status(enum.Enum):
    Stopped = 0
    Opening = 1
    Closing = 2


class Polarity(enum.Enum):
    Positive = 0
    Reverse = 1


class CurtainStatus(DeviceStatus):
    def __init__(self, data: Dict[str, Any]) -> None:
        """Response from device.

        {'id': 1, 'result': [
            {'did': 'current_position', 'siid': 2, 'piid': 3, 'code': 0, 'value': 0},
            {'did': 'status', 'siid': 2, 'piid': 6, 'code': 0, 'value': 0},
            {'did': 'target_position', 'siid': 2, 'piid': 7, 'code': 0, 'value': 0},
            {'did': 'is_manual_enabled', 'siid': 4, 'piid': 1, 'code': 0, 'value': 1},
            {'did': 'polarity', 'siid': 4, 'piid': 2, 'code': 0, 'value': 0},
            {'did': 'is_position_limited', 'siid': 4, 'piid': 3, 'code': 0, 'value': 0},
            {'did': 'night_tip_light', 'siid': 4, 'piid': 4, 'code': 0, 'value': 1},
            {'did': 'run_time', 'siid': 4, 'piid': 5, 'code': 0, 'value': 0},
            {'did': 'adjust_value', 'siid': 5, 'piid': 1, 'code': -4000}
        ]}
        """
        self.data = data

    @property
    def status(self) -> Status:
        """Device status."""
        return Status(self.data["status"])

    @property
    def is_manual_enabled(self) -> bool:
        """True if manual controls are enabled."""
        return bool(self.data["is_manual_enabled"])

    @property
    def polarity(self) -> Polarity:
        """Motor rotation polarity."""
        return Polarity(self.data["polarity"])

    @property
    def is_position_limited(self) -> bool:
        """Position limit."""
        return bool(self.data["is_position_limited"])

    @property
    def night_tip_light(self) -> bool:
        """Night tip light status."""
        return bool(self.data["night_tip_light"])

    @property
    def run_time(self) -> int:
        """Run time of the motor."""
        return self.data["run_time"]

    @property
    def current_position(self) -> int:
        """Current curtain position."""
        return self.data["current_position"]

    @property
    def target_position(self) -> int:
        """Target curtain position."""
        return self.data["target_position"]

    @property
    def adjust_value(self) -> int:
        """Adjust value."""
        return self.data["adjust_value"]


class CurtainMiot(MiotDevice):
    """Main class representing the lumi.curtain.hagl05 curtain."""

    mapping = _MAPPING

    @command(
        default_output=format_output(
            "",
            "Device status: {result.status}\n"
            "Manual enabled: {result.is_manual_enabled}\n"
            "Motor polarity: {result.polarity}\n"
            "Position limit: {result.is_position_limited}\n"
            "Enabled night tip light: {result.night_tip_light}\n"
            "Run time: {result.run_time}\n"
            "Current position: {result.current_position}\n"
            "Target position: {result.target_position}\n"
            "Adjust value: {result.adjust_value}\n",
        )
    )
    def status(self) -> CurtainStatus:
        """Retrieve properties."""

        return CurtainStatus(
            {
                prop["did"]: prop["value"] if prop["code"] == 0 else None
                for prop in self.get_properties_for_mapping()
            }
        )

    @command(
        click.argument("motor_control", type=EnumType(MotorControl)),
        default_output=format_output("Set motor control to {motor_control}"),
    )
    def set_motor_control(self, motor_control: MotorControl):
        """Set motor control."""
        return self.set_property("motor_control", motor_control.value)

    @command(
        click.argument("target_position", type=int),
        default_output=format_output("Set target position to {target_position}"),
    )
    def set_target_position(self, target_position: int):
        """Set target position."""
        if target_position < 0 or target_position > 100:
            raise ValueError(
                "Value must be between [0, 100] value, was %s" % target_position
            )
        return self.set_property("target_position", target_position)

    @command(
        click.argument("manual_enabled", type=bool),
        default_output=format_output("Set manual control {manual_enabled}"),
    )
    def set_manual_enabled(self, manual_enabled: bool):
        """Set manual control of curtain."""
        return self.set_property("is_manual_enabled", manual_enabled)

    @command(
        click.argument("polarity", type=EnumType(Polarity)),
        default_output=format_output("Set polarity to {polarity}"),
    )
    def set_polarity(self, polarity: Polarity):
        """Set polarity of the motor."""
        return self.set_property("polarity", polarity.value)

    @command(
        click.argument("pos_limit", type=bool),
        default_output=format_output("Set position limit to {pos_limit}"),
    )
    def set_position_limit(self, pos_limit: bool):
        """Set position limit parameter."""
        return self.set_property("is_position_limited", pos_limit)

    @command(
        click.argument("night_tip_light", type=bool),
        default_output=format_output("Setting night tip light {night_tip_light"),
    )
    def set_night_tip_light(self, night_tip_light: bool):
        """Set night tip light."""
        return self.set_property("night_tip_light", night_tip_light)

    @command(
        click.argument("adjust_value", type=int),
        default_output=format_output("Set adjust value to {adjust_value}"),
    )
    def set_adjust_value(self, adjust_value: int):
        """Adjust to preferred position."""
        if adjust_value < -100 or adjust_value > 100:
            raise ValueError(
                "Value must be between [-100, 100] value, was %s" % adjust_value
            )
        return self.set_property("adjust_value", adjust_value)



import enum
import re
from typing import Dict, Optional


class FilterType(enum.Enum):
    Regular = "regular"
    AntiBacterial = "anti-bacterial"
    AntiFormaldehyde = "anti-formaldehyde"
    Unknown = "unknown"


FILTER_TYPE_RE = (
    (re.compile(r"^\d+:\d+:41:30$"), FilterType.AntiBacterial),
    (re.compile(r"^\d+:\d+:(30|0|00):31$"), FilterType.AntiFormaldehyde),
    (re.compile(r".*"), FilterType.Regular),
)


class FilterTypeUtil:
    """Utility class for determining xiaomi air filter type."""

    _filter_type_cache: Dict[str, Optional[FilterType]] = {}

    def determine_filter_type(
        self, rfid_tag: Optional[str], product_id: Optional[str]
    ) -> Optional[FilterType]:
        """Determine Xiaomi air filter type based on its product ID.

        :param rfid_tag: RFID tag value
        :param product_id: Product ID such as "0:0:30:33"
        """
        if rfid_tag is None:
            return None
        if rfid_tag == "0:0:0:0:0:0:0":
            return FilterType.Unknown
        if product_id is None:
            return FilterType.Regular

        ft = self._filter_type_cache.get(product_id)
        if ft is None:
            for filter_re, filter_type in FILTER_TYPE_RE:
                if filter_re.match(product_id):
                    ft = self._filter_type_cache[product_id] = filter_type
                    break
        return ft



import enum
import logging
from collections import defaultdict
from typing import Any, Dict, Optional

import click

from .click_common import EnumType, command, format_output
from .device import Device, DeviceStatus
from .exceptions import DeviceException

_LOGGER = logging.getLogger(__name__)

MODEL_AIRDOG_X3 = "airdog.airpurifier.x3"
MODEL_AIRDOG_X5 = "airdog.airpurifier.x5"
MODEL_AIRDOG_X7SM = "airdog.airpurifier.x7sm"

MODEL_AIRDOG_COMMON = ["power", "mode", "speed", "lock", "clean", "pm"]

AVAILABLE_PROPERTIES = {
    MODEL_AIRDOG_X3: MODEL_AIRDOG_COMMON,
    MODEL_AIRDOG_X5: MODEL_AIRDOG_COMMON,
    MODEL_AIRDOG_X7SM: MODEL_AIRDOG_COMMON + ["hcho"],
}


class AirDogException(DeviceException):
    pass


class OperationMode(enum.Enum):
    Auto = "auto"
    Manual = "manual"
    Idle = "sleep"


class OperationModeMapping(enum.Enum):
    Auto = 0
    Manual = 1
    Idle = 2


class AirDogStatus(DeviceStatus):
    """Container for status reports from the air dog x3."""

    def __init__(self, data: Dict[str, Any]) -> None:
        """Response of a Air Dog X3 (airdog.airpurifier.x3):

        {'power: 'on', 'mode': 'sleep', 'speed': 1, 'lock': 'unlock',
         'clean': 'n', 'pm': 11, 'hcho': 0}
        """

        self.data = data

    @property
    def power(self) -> str:
        """Power state."""
        return self.data["power"]

    @property
    def is_on(self) -> bool:
        """True if device is turned on."""
        return self.power == "on"

    @property
    def mode(self) -> OperationMode:
        """Operation mode.

        Can be either auto, manual, sleep.
        """
        return OperationMode(self.data["mode"])

    @property
    def speed(self) -> int:
        """Current speed level."""
        return self.data["speed"]

    @property
    def child_lock(self) -> bool:
        """Return True if child lock is on."""
        return self.data["lock"] == "lock"

    @property
    def clean_filters(self) -> bool:
        """True if the display shows "-C-" and the filter must be cleaned."""
        return self.data["clean"] == "y"

    @property
    def pm25(self) -> int:
        """Return particulate matter value (0...300μg/m³)."""
        return self.data["pm"]

    @property
    def hcho(self) -> Optional[int]:
        """Return formaldehyde value."""
        if self.data["hcho"] is not None:
            return self.data["hcho"]

        return None


class AirDogX3(Device):
    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
        model: str = MODEL_AIRDOG_X3,
    ) -> None:
        super().__init__(ip, token, start_id, debug, lazy_discover)

        if model in AVAILABLE_PROPERTIES:
            self.model = model
        else:
            self.model = MODEL_AIRDOG_X3

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "Mode: {result.mode}\n"
            "Speed: {result.speed}\n"
            "Child lock: {result.child_lock}\n"
            "Clean filters: {result.clean_filters}\n"
            "PM2.5: {result.pm25}\n"
            "Formaldehyde: {result.hcho}\n",
        )
    )
    def status(self) -> AirDogStatus:
        """Retrieve properties."""

        properties = AVAILABLE_PROPERTIES[self.model]
        values = self.get_properties(properties, max_properties=10)

        return AirDogStatus(defaultdict(lambda: None, zip(properties, values)))

    @command(default_output=format_output("Powering on"))
    def on(self):
        """Power on."""
        return self.send("set_power", [1])

    @command(default_output=format_output("Powering off"))
    def off(self):
        """Power off."""
        return self.send("set_power", [0])

    @command(
        click.argument("mode", type=EnumType(OperationMode)),
        click.argument("speed", type=int, required=False, default=1),
        default_output=format_output(
            "Setting mode to '{mode.value}' and speed to {speed}"
        ),
    )
    def set_mode_and_speed(self, mode: OperationMode, speed: int = 1):
        """Set mode and speed."""
        if mode.value not in (om.value for om in OperationMode):
            raise AirDogException(
                "{} is not a valid OperationMode value".format(mode.value)
            )

        if mode in [OperationMode.Auto, OperationMode.Idle]:
            speed = 1

        if self.model == MODEL_AIRDOG_X3:
            max_speed = 4
        else:
            # airdog.airpurifier.x7, airdog.airpurifier.x7sm
            max_speed = 5

        if speed < 1 or speed > max_speed:
            raise AirDogException("Invalid speed: %s" % speed)

        return self.send("set_wind", [OperationModeMapping[mode.name].value, speed])

    @command(
        click.argument("lock", type=bool),
        default_output=format_output(
            lambda lock: "Turning on child lock" if lock else "Turning off child lock"
        ),
    )
    def set_child_lock(self, lock: bool):
        """Set child lock on/off."""
        return self.send("set_lock", [int(lock)])

    @command(default_output=format_output("Setting filters cleaned"))
    def set_filters_cleaned(self):
        """Set filters cleaned."""
        return self.send("set_clean")


class AirDogX5(AirDogX3):
    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
        model: str = MODEL_AIRDOG_X5,
    ) -> None:
        super().__init__(ip, token, start_id, debug, lazy_discover)

        if model in AVAILABLE_PROPERTIES:
            self.model = model
        else:
            self.model = MODEL_AIRDOG_X5


class AirDogX7SM(AirDogX3):
    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
        model: str = MODEL_AIRDOG_X7SM,
    ) -> None:
        super().__init__(ip, token, start_id, debug, lazy_discover)

        if model in AVAILABLE_PROPERTIES:
            self.model = model
        else:
            self.model = MODEL_AIRDOG_X7SM



import enum
import logging
from collections import defaultdict
from typing import Any, Dict, Optional

import click

from .airfilter_util import FilterType, FilterTypeUtil
from .click_common import EnumType, command, format_output
from .device import Device, DeviceStatus
from .exceptions import DeviceException

_LOGGER = logging.getLogger(__name__)


class AirPurifierException(DeviceException):
    pass


class OperationMode(enum.Enum):
    # Supported modes of the Air Purifier Pro, 2, V3
    Auto = "auto"
    Silent = "silent"
    Favorite = "favorite"
    # Additional supported modes of the Air Purifier 2 and V3
    Idle = "idle"
    # Additional supported modes of the Air Purifier V3
    Medium = "medium"
    High = "high"
    Strong = "strong"
    # Additional supported modes of the Air Purifier Super 2
    Low = "low"


class SleepMode(enum.Enum):
    Off = "poweroff"
    Silent = "silent"
    Idle = "idle"


class LedBrightness(enum.Enum):
    Bright = 0
    Dim = 1
    Off = 2


class AirPurifierStatus(DeviceStatus):
    """Container for status reports from the air purifier."""

    def __init__(self, data: Dict[str, Any]) -> None:
        """Response of a Air Purifier Pro (zhimi.airpurifier.v6):

        {'power': 'off', 'aqi': 7, 'average_aqi': 18, 'humidity': 45,
         'temp_dec': 234, 'mode': 'auto', 'favorite_level': 17,
         'filter1_life': 52, 'f1_hour_used': 1664, 'use_time': 2642700,
         'motor1_speed': 0, 'motor2_speed': 800, 'purify_volume': 62180,
         'f1_hour': 3500, 'led': 'on', 'led_b': None, 'bright': 83,
         'buzzer': None, 'child_lock': 'off', 'volume': 50,
         'rfid_product_id': '0:0:41:30', 'rfid_tag': '80:52:86:e2:d8:86:4',
         'act_sleep': 'close'}

        Response of a Air Purifier Pro (zhimi.airpurifier.v7):

        {'power': 'on', 'aqi': 2, 'average_aqi': 3, 'humidity': 42,
         'temp_dec': 223, 'mode': 'favorite', 'favorite_level': 3,
         'filter1_life': 56, 'f1_hour_used': 1538, 'use_time': None,
         'motor1_speed': 300, 'motor2_speed': 898, 'purify_volume': None,
         'f1_hour': 3500, 'led': 'on', 'led_b': None, 'bright': 45,
         'buzzer': None, 'child_lock': 'off', 'volume': 0,
         'rfid_product_id': '0:0:30:33', 'rfid_tag': '80:6a:a9:e2:37:92:4',
         'act_sleep': None, 'sleep_mode': None, 'sleep_time': None,
         'sleep_data_num': None, 'app_extra': 0, 'act_det': None,
         'button_pressed': None}

        Response of a Air Purifier 2 (zhimi.airpurifier.m1):

        {'power': 'on, 'aqi': 10, 'average_aqi': 8, 'humidity': 62,
         'temp_dec': 186, 'mode': 'auto', 'favorite_level': 10,
        'filter1_life': 80, 'f1_hour_used': 682, 'use_time': 2457000,
        'motor1_speed': 354, 'motor2_speed': None, 'purify_volume': 25262,
        'f1_hour': 3500, 'led': 'off', 'led_b': 2, 'bright': None,
        'buzzer': 'off', 'child_lock': 'off', 'volume': None,
        'rfid_product_id': None, 'rfid_tag': None,
        'act_sleep': 'close'}

        Response of a Air Purifier 2 (zhimi.airpurifier.m2):

        {'power': 'on', 'aqi': 10, 'average_aqi': 8, 'humidity': 42,
         'temp_dec': 223, 'mode': 'favorite', 'favorite_level': 2,
         'filter1_life': 63, 'f1_hour_used': 1282, 'use_time': 16361416,
         'motor1_speed': 747, 'motor2_speed': None, 'purify_volume': 421580,
         'f1_hour': 3500, 'led': 'on', 'led_b': 1, 'bright': None,
         'buzzer': 'off', 'child_lock': 'off', 'volume': None,
         'rfid_product_id': None, 'rfid_tag': None, 'act_sleep': 'close',
         'sleep_mode': 'idle', 'sleep_time': 86168, 'sleep_data_num': 30,
         'app_extra': 0, 'act_det': None, 'button_pressed': None}

        Response of a Air Purifier V3 (zhimi.airpurifier.v3)

        {'power': 'off', 'aqi': 0, 'humidity': None, 'temp_dec': None,
         'mode': 'idle', 'led': 'off', 'led_b': 10, 'buzzer': 'on',
         'child_lock': 'off', 'bright': 43, 'favorite_level': None,
         'filter1_life': 26, 'f1_hour_used': 2573, 'use_time': None,
         'motor1_speed': 0}

        {'power': 'on', 'aqi': 18, 'humidity': None, 'temp_dec': None,
         'mode': 'silent', 'led': 'off', 'led_b': 10, 'buzzer': 'on',
         'child_lock': 'off', 'bright': 4, 'favorite_level': None,
         'filter1_life': 26, 'f1_hour_used': 2574, 'use_time': None,
         'motor1_speed': 648}

        A request is limited to 16 properties.
        """

        self.filter_type_util = FilterTypeUtil()
        self.data = data

    @property
    def power(self) -> str:
        """Power state."""
        return self.data["power"]

    @property
    def is_on(self) -> bool:
        """Return True if device is on."""
        return self.power == "on"

    @property
    def aqi(self) -> int:
        """Air quality index."""
        return self.data["aqi"]

    @property
    def average_aqi(self) -> int:
        """Average of the air quality index."""
        return self.data["average_aqi"]

    @property
    def humidity(self) -> int:
        """Current humidity."""
        return self.data["humidity"]

    @property
    def temperature(self) -> Optional[float]:
        """Current temperature, if available."""
        if self.data["temp_dec"] is not None:
            return self.data["temp_dec"] / 10.0

        return None

    @property
    def mode(self) -> OperationMode:
        """Current operation mode."""
        return OperationMode(self.data["mode"])

    @property
    def sleep_mode(self) -> Optional[SleepMode]:
        """Operation mode of the sleep state.

        (Idle vs. Silent)
        """
        if self.data["sleep_mode"] is not None:
            return SleepMode(self.data["sleep_mode"])

        return None

    @property
    def led(self) -> bool:
        """Return True if LED is on."""
        return self.data["led"] == "on"

    @property
    def led_brightness(self) -> Optional[LedBrightness]:
        """Brightness of the LED."""
        if self.data["led_b"] is not None:
            try:
                return LedBrightness(self.data["led_b"])
            except ValueError:
                return None

        return None

    @property
    def illuminance(self) -> Optional[int]:
        """Environment illuminance level in lux [0-200].

        Sensor value is updated only when device is turned on.
        """
        return self.data["bright"]

    @property
    def buzzer(self) -> Optional[bool]:
        """Return True if buzzer is on."""
        if self.data["buzzer"] is not None:
            return self.data["buzzer"] == "on"

        return None

    @property
    def child_lock(self) -> bool:
        """Return True if child lock is on."""
        return self.data["child_lock"] == "on"

    @property
    def favorite_level(self) -> int:
        """Return favorite level, which is used if the mode is ``favorite``."""
        # Favorite level used when the mode is `favorite`.
        return self.data["favorite_level"]

    @property
    def filter_life_remaining(self) -> int:
        """Time until the filter should be changed."""
        return self.data["filter1_life"]

    @property
    def filter_hours_used(self) -> int:
        """How long the filter has been in use."""
        return self.data["f1_hour_used"]

    @property
    def use_time(self) -> int:
        """How long the device has been active in seconds."""
        return self.data["use_time"]

    @property
    def purify_volume(self) -> int:
        """The volume of purified air in cubic meter."""
        return self.data["purify_volume"]

    @property
    def motor_speed(self) -> int:
        """Speed of the motor."""
        return self.data["motor1_speed"]

    @property
    def motor2_speed(self) -> Optional[int]:
        """Speed of the 2nd motor."""
        return self.data["motor2_speed"]

    @property
    def volume(self) -> Optional[int]:
        """Volume of sound notifications [0-100]."""
        return self.data["volume"]

    @property
    def filter_rfid_product_id(self) -> Optional[str]:
        """RFID product ID of installed filter."""
        return self.data["rfid_product_id"]

    @property
    def filter_rfid_tag(self) -> Optional[str]:
        """RFID tag ID of installed filter."""
        return self.data["rfid_tag"]

    @property
    def filter_type(self) -> Optional[FilterType]:
        """Type of installed filter."""
        return self.filter_type_util.determine_filter_type(
            self.filter_rfid_tag, self.filter_rfid_product_id
        )

    @property
    def learn_mode(self) -> bool:
        """Return True if Learn Mode is enabled."""
        return self.data["act_sleep"] == "single"

    @property
    def sleep_time(self) -> Optional[int]:
        return self.data["sleep_time"]

    @property
    def sleep_mode_learn_count(self) -> Optional[int]:
        return self.data["sleep_data_num"]

    @property
    def extra_features(self) -> Optional[int]:
        return self.data["app_extra"]

    @property
    def turbo_mode_supported(self) -> Optional[bool]:
        if self.data["app_extra"] is not None:
            return self.data["app_extra"] == 1

        return None

    @property
    def auto_detect(self) -> Optional[bool]:
        """Return True if auto detect is enabled."""
        if self.data["act_det"] is not None:
            return self.data["act_det"] == "on"

        return None

    @property
    def button_pressed(self) -> Optional[str]:
        """Last pressed button."""
        return self.data["button_pressed"]


class AirPurifier(Device):
    """Main class representing the air purifier."""

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "AQI: {result.aqi} μg/m³\n"
            "Average AQI: {result.average_aqi} μg/m³\n"
            "Temperature: {result.temperature} °C\n"
            "Humidity: {result.humidity} %\n"
            "Mode: {result.mode.value}\n"
            "LED: {result.led}\n"
            "LED brightness: {result.led_brightness}\n"
            "Illuminance: {result.illuminance} lx\n"
            "Buzzer: {result.buzzer}\n"
            "Child lock: {result.child_lock}\n"
            "Favorite level: {result.favorite_level}\n"
            "Filter life remaining: {result.filter_life_remaining} %\n"
            "Filter hours used: {result.filter_hours_used}\n"
            "Use time: {result.use_time} s\n"
            "Purify volume: {result.purify_volume} m³\n"
            "Motor speed: {result.motor_speed} rpm\n"
            "Motor 2 speed: {result.motor2_speed} rpm\n"
            "Sound volume: {result.volume} %\n"
            "Filter RFID product id: {result.filter_rfid_product_id}\n"
            "Filter RFID tag: {result.filter_rfid_tag}\n"
            "Filter type: {result.filter_type}\n"
            "Learn mode: {result.learn_mode}\n"
            "Sleep mode: {result.sleep_mode}\n"
            "Sleep time: {result.sleep_time}\n"
            "Sleep mode learn count: {result.sleep_mode_learn_count}\n"
            "AQI sensor enabled on power off: {result.auto_detect}\n",
        )
    )
    def status(self) -> AirPurifierStatus:
        """Retrieve properties."""

        properties = [
            "power",
            "aqi",
            "average_aqi",
            "humidity",
            "temp_dec",
            "mode",
            "favorite_level",
            "filter1_life",
            "f1_hour_used",
            "use_time",
            "motor1_speed",
            "motor2_speed",
            "purify_volume",
            "f1_hour",
            "led",
            # Second request
            "led_b",
            "bright",
            "buzzer",
            "child_lock",
            "volume",
            "rfid_product_id",
            "rfid_tag",
            "act_sleep",
            "sleep_mode",
            "sleep_time",
            "sleep_data_num",
            "app_extra",
            "act_det",
            "button_pressed",
        ]

        values = self.get_properties(properties, max_properties=15)

        return AirPurifierStatus(defaultdict(lambda: None, zip(properties, values)))

    @command(default_output=format_output("Powering on"))
    def on(self):
        """Power on."""
        return self.send("set_power", ["on"])

    @command(default_output=format_output("Powering off"))
    def off(self):
        """Power off."""
        return self.send("set_power", ["off"])

    @command(
        click.argument("mode", type=EnumType(OperationMode)),
        default_output=format_output("Setting mode to '{mode.value}'"),
    )
    def set_mode(self, mode: OperationMode):
        """Set mode."""
        return self.send("set_mode", [mode.value])

    @command(
        click.argument("level", type=int),
        default_output=format_output("Setting favorite level to {level}"),
    )
    def set_favorite_level(self, level: int):
        """Set favorite level."""
        if level < 0 or level > 17:
            raise AirPurifierException("Invalid favorite level: %s" % level)

        # Possible alternative property: set_speed_favorite

        # Set the favorite level used when the mode is `favorite`,
        # should be  between 0 and 17.
        return self.send("set_level_favorite", [level])  # 0 ... 17

    @command(
        click.argument("brightness", type=EnumType(LedBrightness)),
        default_output=format_output("Setting LED brightness to {brightness}"),
    )
    def set_led_brightness(self, brightness: LedBrightness):
        """Set led brightness."""
        return self.send("set_led_b", [brightness.value])

    @command(
        click.argument("led", type=bool),
        default_output=format_output(
            lambda led: "Turning on LED" if led else "Turning off LED"
        ),
    )
    def set_led(self, led: bool):
        """Turn led on/off."""
        if led:
            return self.send("set_led", ["on"])
        else:
            return self.send("set_led", ["off"])

    @command(
        click.argument("buzzer", type=bool),
        default_output=format_output(
            lambda buzzer: "Turning on buzzer" if buzzer else "Turning off buzzer"
        ),
    )
    def set_buzzer(self, buzzer: bool):
        """Set buzzer on/off."""
        if buzzer:
            return self.send("set_buzzer", ["on"])
        else:
            return self.send("set_buzzer", ["off"])

    @command(
        click.argument("lock", type=bool),
        default_output=format_output(
            lambda lock: "Turning on child lock" if lock else "Turning off child lock"
        ),
    )
    def set_child_lock(self, lock: bool):
        """Set child lock on/off."""
        if lock:
            return self.send("set_child_lock", ["on"])
        else:
            return self.send("set_child_lock", ["off"])

    @command(
        click.argument("volume", type=int),
        default_output=format_output("Setting sound volume to {volume}"),
    )
    def set_volume(self, volume: int):
        """Set volume of sound notifications [0-100]."""
        if volume < 0 or volume > 100:
            raise AirPurifierException("Invalid volume: %s" % volume)

        return self.send("set_volume", [volume])

    @command(
        click.argument("learn_mode", type=bool),
        default_output=format_output(
            lambda learn_mode: "Turning on learn mode"
            if learn_mode
            else "Turning off learn mode"
        ),
    )
    def set_learn_mode(self, learn_mode: bool):
        """Set the Learn Mode on/off."""
        if learn_mode:
            return self.send("set_act_sleep", ["single"])
        else:
            return self.send("set_act_sleep", ["close"])

    @command(
        click.argument("auto_detect", type=bool),
        default_output=format_output(
            lambda auto_detect: "Turning on auto detect"
            if auto_detect
            else "Turning off auto detect"
        ),
    )
    def set_auto_detect(self, auto_detect: bool):
        """Set auto detect on/off.

        It's a feature of the AirPurifier V1 & V3
        """
        if auto_detect:
            return self.send("set_act_det", ["on"])
        else:
            return self.send("set_act_det", ["off"])

    @command(
        click.argument("value", type=int),
        default_output=format_output("Setting extra to {value}"),
    )
    def set_extra_features(self, value: int):
        """Storage register to enable extra features at the app.

        app_extra=1 unlocks a turbo mode supported feature
        """
        if value < 0:
            raise AirPurifierException("Invalid app extra value: %s" % value)

        return self.send("set_app_extra", [value])

    @command(default_output=format_output("Resetting filter"))
    def reset_filter(self):
        """Resets filter hours used and remaining life."""
        return self.send("reset_filter1")



import ast
import logging
import sys
from typing import Any  # noqa: F401

import click

import miio  # noqa: E402
from miio.click_common import ExceptionHandlerGroup, validate_ip, validate_token
from miio.miioprotocol import MiIOProtocol

_LOGGER = logging.getLogger(__name__)
pass_dev = click.make_pass_decorator(miio.ChuangmiPlug)


@click.group(invoke_without_command=True, cls=ExceptionHandlerGroup)
@click.option("--ip", envvar="DEVICE_IP", callback=validate_ip)
@click.option("--token", envvar="DEVICE_TOKEN", callback=validate_token)
@click.option("-d", "--debug", default=False, count=True)
@click.pass_context
def cli(ctx, ip: str, token: str, debug: int):
    """A tool to command Xiaomi Smart Plug."""
    if debug:
        logging.basicConfig(level=logging.DEBUG)
        _LOGGER.info("Debug mode active")
    else:
        logging.basicConfig(level=logging.INFO)

    _LOGGER.warning(
        "This script is deprecated and will be removed soon, use `miiocli chuangmiplug` instead"
    )

    # if we are scanning, we do not try to connect.
    if ctx.invoked_subcommand == "discover":
        return

    if ip is None or token is None:
        click.echo("You have to give ip and token!")
        sys.exit(-1)

    dev = miio.ChuangmiPlug(ip, token, debug)
    _LOGGER.debug("Connecting to %s with token %s", ip, token)

    ctx.obj = dev

    if ctx.invoked_subcommand is None:
        ctx.invoke(status)


@cli.command()
def discover():
    """Search for plugs in the network."""
    MiIOProtocol.discover()


@cli.command()
@pass_dev
def status(dev: miio.ChuangmiPlug):
    """Returns the state information."""
    res = dev.status()
    if not res:
        return  # bail out

    click.echo(click.style("Power: %s" % res.power, bold=True))
    click.echo("Temperature: %s" % res.temperature)


@cli.command()
@pass_dev
def on(dev: miio.ChuangmiPlug):
    """Power on."""
    click.echo("Power on: %s" % dev.on())


@cli.command()
@pass_dev
def off(dev: miio.ChuangmiPlug):
    """Power off."""
    click.echo("Power off: %s" % dev.off())


@cli.command()
@click.argument("cmd", required=True)
@click.argument("parameters", required=False)
@pass_dev
def raw_command(dev: miio.ChuangmiPlug, cmd, parameters):
    """Run a raw command."""
    params = []  # type: Any
    if parameters:
        params = ast.literal_eval(parameters)
    click.echo("Sending cmd %s with params %s" % (cmd, params))
    click.echo(dev.raw_command(cmd, params))


if __name__ == "__main__":
    cli()



import enum
import time

import click

from .click_common import EnumType, command
from .device import Device, DeviceStatus


class HourlySystem(enum.Enum):
    TwentyFour = 24
    Twelve = 12


class AlarmType(enum.Enum):
    Alarm = "alarm"
    Reminder = "reminder"
    Timer = "timer"


# TODO names for the tones
class Tone(enum.Enum):
    First = "a1.mp3"
    Second = "a2.mp3"
    Third = "a3.mp3"
    Fourth = "a4.mp3"
    Fifth = "a5.mp3"
    Sixth = "a6.mp3"
    Seventh = "a7.mp3"


class Nightmode(DeviceStatus):
    def __init__(self, data):
        self._enabled = bool(data[0])
        self._start = data[1]
        self._end = data[2]

    @property
    def enabled(self) -> bool:
        return self._enabled

    @property
    def start(self):
        return self._start

    @property
    def end(self):
        return self._end


class RingTone(DeviceStatus):
    def __init__(self, data):
        # {'type': 'reminder', 'ringtone': 'a2.mp3', 'smart_clock': 0}]
        self.type = AlarmType(data["type"])
        self.tone = Tone(data["ringtone"])
        self.smart_clock = data["smart_clock"]


class AlarmClock(Device):
    """Implementation of Xiao AI Smart Alarm Clock.

    Note, this device is not very responsive to the requests, so it may take several
    seconds /tries to get an answer.
    """

    @command()
    def get_config_version(self):
        """
        # values unknown {'result': [4], 'id': 203}
        :return:
        """
        return self.send("get_config_version", ["audio"])

    @command()
    def clock_system(self) -> HourlySystem:
        """Returns either 12 or 24 depending on which system is in use."""
        return HourlySystem(self.send("get_hourly_system")[0])

    @command(click.argument("brightness", type=EnumType(HourlySystem)))
    def set_hourly_system(self, hs: HourlySystem):
        return self.send("set_hourly_system", [hs.value])

    @command()
    def get_button_light(self):
        """Get button's light state."""
        # ['normal', 'mute', 'offline'] or []
        return self.send("get_enabled_key_light")

    @command(click.argument("on", type=bool))
    def set_button_light(self, on):
        """Enable or disable the button light."""
        if on:
            return self.send("enable_key_light") == ["OK"]
        else:
            return self.send("disable_key_light") == ["OK"]

    @command()
    def volume(self) -> int:
        """Return the volume.

        ->  192.168.0.128 data= {"id":251,"method":"set_volume","params":[17]}
        <-  192.168.0.57 data= {"result":["OK"],"id":251}
        """
        return int(self.send("get_volume")[0])

    @command(click.argument("volume", type=int))
    def set_volume(self, volume):
        """Set volume [1,100]."""
        return self.send("set_volume", [volume]) == ["OK"]

    @command(
        click.argument(
            "alarm_type", type=EnumType(AlarmType), default=AlarmType.Alarm.name
        )
    )
    def get_ring(self, alarm_type: AlarmType):
        """Get current ring tone settings."""
        return RingTone(self.send("get_ring", [{"type": alarm_type.value}]).pop())

    @command(
        click.argument("alarm_type", type=EnumType(AlarmType)),
        click.argument("tone", type=EnumType(Tone)),
    )
    def set_ring(self, alarm_type: AlarmType, ring: RingTone):
        """Set alarm tone.

        ->  192.168.0.128 data= {"id":236,"method":"set_ring",
           "params":[{"ringtone":"a1.mp3","smart_clock":"","type":"alarm"}]}
        <-  192.168.0.57 data= {"result":["OK"],"id":236}
        """
        raise NotImplementedError()
        # return self.send("set_ring", ) == ["OK"]

    @command()
    def night_mode(self):
        """Get night mode status.

        ->  192.168.0.128 data= {"id":234,"method":"get_night_mode","params":[]}
        <-  192.168.0.57 data= {"result":[0],"id":234}
        """
        return Nightmode(self.send("get_night_mode"))

    @command()
    def set_night_mode(self):
        """Set the night mode.

        # enable
        ->  192.168.0.128 data= {"id":248,"method":"set_night_mode",
            "params":[1,"21:00","6:00"]}
        <-  192.168.0.57 data= {"result":["OK"],"id":248}

        # disable
        ->  192.168.0.128 data= {"id":249,"method":"set_night_mode",
            "params":[0,"21:00","6:00"]}
        <-  192.168.0.57 data= {"result":["OK"],"id":249}
        """
        raise NotImplementedError()

    @command()
    def near_wakeup(self):
        """Status for near wakeup.

        ->  192.168.0.128 data= {"id":235,"method":"get_near_wakeup_status",
            "params":[]}
        <-  192.168.0.57 data= {"result":["disable"],"id":235}

        # setters
        ->  192.168.0.128 data= {"id":254,"method":"set_near_wakeup_status",
            "params":["enable"]}
        <-  192.168.0.57 data= {"result":["OK"],"id":254}

        ->  192.168.0.128 data= {"id":255,"method":"set_near_wakeup_status",
            "params":["disable"]}
        <-  192.168.0.57 data= {"result":["OK"],"id":255}
        """
        return self.send("get_near_wakeup_status")

    @command()
    def countdown(self):
        """
        ->  192.168.0.128 data= {"id":258,"method":"get_count_down_v2","params":[]}
        """
        return self.send("get_count_down_v2")

    @command()
    def alarmops(self):
        """
        NOTE: the alarm_ops method is the one used to create, query and delete
        all types of alarms (reminders, alarms, countdowns).
        ->  192.168.0.128 data= {"id":263,"method":"alarm_ops",
            "params":{"operation":"create","data":[
                {"type":"alarm","event":"testlabel","reminder":"","smart_clock":0,
                "ringtone":"a2.mp3","volume":100,"circle":"once","status":"on",
                "repeat_ringing":0,"delete_datetime":1564291980000,
                "disable_datetime":"","circle_extra":"",
                "datetime":1564291980000}
                ],"update_datetime":1564205639326}}
        <-  192.168.0.57 data= {"result":[{"id":1,"ack":"OK"}],"id":263}

        # query per index, starts from 0 instead of 1 as the ids it seems
        ->  192.168.0.128 data= {"id":264,"method":"alarm_ops",
            "params":{"operation":"query","req_type":"alarm",
                "update_datetime":1564205639593,"index":0}}
        <-  192.168.0.57 data= {"result":
            [0,[
                {"i":"1","c":"once","d":"2019-07-28T13:33:00+0800","s":"on",
                "n":"testlabel","a":"a2.mp3","dd":1}
               ], "America/New_York"
            ],"id":264}

        # result [code, list of alarms, timezone]
        ->  192.168.0.128 data= {"id":265,"method":"alarm_ops",
            "params":{"operation":"query","index":0,"update_datetime":1564205639596,
                "req_type":"reminder"}}
        <-  192.168.0.57 data= {"result":[0,[],"America/New_York"],"id":265}
        """
        raise NotImplementedError()

    @command(click.argument("url"))
    def start_countdown(self, url):
        """Start countdown timer playing the given media.

        {"id":354,"method":"alarm_ops",
        "params":{"operation":"create","update_datetime":1564206432733,
        "data":[{"type":"timer",
          "background":"http://host.invalid/testfile.mp3",
          "offset":1800,
          "circle":"once",
          "volume":100,
          "datetime":1564208232733}]}}
        """

        current_ts = int(time.time() * 1000)
        payload = {
            "operation": "create",
            "update_datetime": current_ts,
            "data": [
                {
                    "type": "timer",
                    "background": "http://url_here_for_mp3",
                    "offset": 30,
                    "circle": "once",
                    "volume": 30,
                    "datetime": current_ts,
                }
            ],
        }

        return self.send("alarm_ops", payload)

    @command()
    def query(self):
        """
        ->  192.168.0.128 data= {"id":227,"method":"alarm_ops","params":
        {"operation":"query","index":0,"update_datetime":1564205198413,"req_type":"reminder"}}

        """

        payload = {
            "operation": "query",
            "index": 0,
            "update_datetime": int(time.time() * 1000),
            "req_type": "timer",
        }
        return self.send("alarm_ops", payload)

    @command()
    def cancel(self):
        """Cancel alarm of the defined type.

        "params":{"operation":"cancel","update_datetime":1564206332603,"data":[{"type":"timer"}]}}
        """
        import time

        payload = {
            "operation": "pause",
            "update_datetime": int(time.time() * 1000),
            "data": [{"type": "timer"}],
        }
        return self.send("alarm_ops", payload)



import hashlib
import logging
from http.server import BaseHTTPRequestHandler, HTTPServer
from os.path import basename

_LOGGER = logging.getLogger(__name__)


class SingleFileHandler(BaseHTTPRequestHandler):
    """A simplified handler just returning the contents of a buffer."""

    def __init__(self, request, client_address, server):
        self.payload = server.payload
        self.server = server

        super().__init__(request, client_address, server)

    def handle_one_request(self):
        self.server.got_request = True
        self.raw_requestline = self.rfile.readline()

        if not self.parse_request():
            _LOGGER.error("unable to parse request: %s" % self.raw_requestline)
            return

        self.send_response(200)
        self.send_header("Content-type", "application/octet-stream")
        self.send_header("Content-Length", len(self.payload))
        self.end_headers()
        self.wfile.write(self.payload)


class OneShotServer:
    """A simple HTTP server for serving an update file.

    The server will be started in an emphemeral port, and will only accept a single
    request to keep it simple.
    """

    def __init__(self, file, interface=None):
        addr = ("", 0)
        self.server = HTTPServer(addr, SingleFileHandler)
        setattr(self.server, "got_request", False)  # noqa: B010

        self.addr, self.port = self.server.server_address
        self.server.timeout = 10

        _LOGGER.info(
            "Serving on %s:%s, timeout %s" % (self.addr, self.port, self.server.timeout)
        )

        self.file = basename(file)
        with open(file, "rb") as f:
            self.payload = f.read()
            self.server.payload = self.payload
            self.md5 = hashlib.md5(self.payload).hexdigest()  # nosec
            _LOGGER.info("Using local %s (md5: %s)" % (file, self.md5))

    @staticmethod
    def find_local_ip():
        try:
            import netifaces
        except Exception:
            _LOGGER.error(
                "Unable to import netifaces, please install netifaces library"
            )
            raise

        ifaces_without_lo = [
            x for x in netifaces.interfaces() if not x.startswith("lo")
        ]
        _LOGGER.debug("available interfaces: %s" % ifaces_without_lo)

        for iface in ifaces_without_lo:
            addresses = netifaces.ifaddresses(iface)
            if netifaces.AF_INET not in addresses:
                _LOGGER.debug("%s has no ipv4 addresses, skipping" % iface)
                continue
            for entry in addresses[netifaces.AF_INET]:
                _LOGGER.debug("Got addr: %s" % entry["addr"])
                return entry["addr"]

    def url(self, ip=None):
        if ip is None:
            ip = OneShotServer.find_local_ip()

        url = "http://%s:%s/%s" % (ip, self.port, self.file)
        return url

    def serve_once(self):
        self.server.handle_request()
        if getattr(self.server, "got_request"):  # noqa: B009
            _LOGGER.info("Got a request, should be downloading now.")
            return True
        else:
            _LOGGER.error("No request was made..")
            return False


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    upd = OneShotServer("/tmp/test")  # nosec
    upd.serve_once()



import binascii

import pytest

from miio.exceptions import DeviceError, PayloadDecodeException, RecoverableError

from .. import Utils
from ..miioprotocol import MiIOProtocol
from ..protocol import Message

METHOD = "method"
PARAMS = "params"


@pytest.fixture
def proto() -> MiIOProtocol:
    return MiIOProtocol()


@pytest.fixture
def token() -> bytes:
    return bytes.fromhex(32 * "0")


def build_msg(data, token):
    encrypted_data = Utils.encrypt(data, token)

    # header
    magic = binascii.unhexlify(b"2131")
    length = (32 + len(encrypted_data)).to_bytes(2, byteorder="big")
    unknown = binascii.unhexlify(b"00000000")
    did = binascii.unhexlify(b"01234567")
    epoch = binascii.unhexlify(b"00000000")

    checksum = Utils.md5(
        magic + length + unknown + did + epoch + token + encrypted_data
    )

    return magic + length + unknown + did + epoch + checksum + encrypted_data


def test_incrementing_id(proto):
    old_id = proto.raw_id
    proto._create_request("dummycmd", "dummy")
    assert proto.raw_id > old_id


def test_id_loop(proto):
    proto.__id = 9999
    proto._create_request("dummycmd", "dummy")
    assert proto.raw_id == 1


def test_request_with_none_param(proto):
    req = proto._create_request("dummy", None)
    assert isinstance(req["params"], list)
    assert len(req["params"]) == 0


def test_request_with_string_param(proto):
    req = proto._create_request("command", "single")
    assert req[METHOD] == "command"
    assert req[PARAMS] == "single"


def test_request_with_list_param(proto):
    req = proto._create_request("command", ["item"])
    assert req[METHOD] == "command"
    assert req[PARAMS] == ["item"]


def test_request_extra_params(proto):
    req = proto._create_request("command", ["item"], extra_parameters={"sid": 1234})
    assert "sid" in req
    assert req["sid"] == 1234


def test_device_error_handling(proto: MiIOProtocol):
    retry_error = -30001
    with pytest.raises(RecoverableError):
        proto._handle_error({"code": retry_error})

    with pytest.raises(DeviceError):
        proto._handle_error({"code": 1234})


def test_non_bytes_payload(token):
    payload = "hello world"
    with pytest.raises(TypeError):
        Utils.encrypt(payload, token)
    with pytest.raises(TypeError):
        Utils.decrypt(payload, token)


def test_encrypt(token):
    payload = b"hello world"

    encrypted = Utils.encrypt(payload, token)
    decrypted = Utils.decrypt(encrypted, token)
    assert payload == decrypted


def test_invalid_token():
    payload = b"hello world"
    wrong_type = 1234
    wrong_length = bytes.fromhex(16 * "0")
    with pytest.raises(TypeError):
        Utils.encrypt(payload, wrong_type)
    with pytest.raises(TypeError):
        Utils.decrypt(payload, wrong_type)

    with pytest.raises(ValueError):
        Utils.encrypt(payload, wrong_length)
    with pytest.raises(ValueError):
        Utils.decrypt(payload, wrong_length)


def test_decode_json_payload(token):
    ctx = {"token": token}

    # can parse message with valid json
    serialized_msg = build_msg(b'{"id": 123456}', token)
    parsed_msg = Message.parse(serialized_msg, **ctx)
    assert parsed_msg.data.value
    assert isinstance(parsed_msg.data.value, dict)
    assert parsed_msg.data.value["id"] == 123456


def test_decode_json_quirk_powerstrip(token):
    ctx = {"token": token}

    # can parse message with invalid json for edge case powerstrip
    # when not connected to cloud
    serialized_msg = build_msg(b'{"id": 123456,,"otu_stat":0}', token)
    parsed_msg = Message.parse(serialized_msg, **ctx)
    assert parsed_msg.data.value
    assert isinstance(parsed_msg.data.value, dict)
    assert parsed_msg.data.value["id"] == 123456
    assert parsed_msg.data.value["otu_stat"] == 0


def test_decode_json_quirk_cloud(token):
    ctx = {"token": token}

    # can parse message with invalid json for edge case xiaomi cloud
    # reply to _sync.batch_gen_room_up_url
    serialized_msg = build_msg(b'{"id": 123456}\x00k', token)
    parsed_msg = Message.parse(serialized_msg, **ctx)
    assert parsed_msg.data.value
    assert isinstance(parsed_msg.data.value, dict)
    assert parsed_msg.data.value["id"] == 123456


def test_decode_json_raises_for_invalid_json(token):
    ctx = {"token": token}

    # make sure PayloadDecodeDexception is raised for invalid json
    serialized_msg = build_msg(b'{"id": 123456,,"otu_stat":0', token)
    with pytest.raises(PayloadDecodeException):
        Message.parse(serialized_msg, **ctx)



from unittest import TestCase

import pytest

from miio import YeelightDualControlModule
from miio.yeelight_dual_switch import Switch, YeelightDualControlModuleException

from .dummies import DummyMiotDevice

_INITIAL_STATE = {
    "switch_1_state": True,
    "switch_1_default_state": True,
    "switch_1_off_delay": 300,
    "switch_2_state": False,
    "switch_2_default_state": False,
    "switch_2_off_delay": 0,
    "interlock": False,
    "flex_mode": True,
    "rc_list": "[{'mac':'9db0eb4124f8','evtid':4097,'pid':339,'beaconkey':'3691bc0679eef9596bb63abf'}]",
}


class DummyYeelightDualControlModule(DummyMiotDevice, YeelightDualControlModule):
    def __init__(self, *args, **kwargs):
        self.state = _INITIAL_STATE
        self.return_values = {
            "get_prop": self._get_state,
        }
        super().__init__(*args, **kwargs)


@pytest.fixture(scope="function")
def switch(request):
    request.cls.device = DummyYeelightDualControlModule()


@pytest.mark.usefixtures("switch")
class TestYeelightDualControlModule(TestCase):
    def test_1_on(self):
        self.device.off(Switch.First)  # ensure off
        assert self.device.status().switch_1_state is False

        self.device.on(Switch.First)
        assert self.device.status().switch_1_state is True

    def test_2_on(self):
        self.device.off(Switch.Second)  # ensure off
        assert self.device.status().switch_2_state is False

        self.device.on(Switch.Second)
        assert self.device.status().switch_2_state is True

    def test_1_off(self):
        self.device.on(Switch.First)  # ensure on
        assert self.device.status().switch_1_state is True

        self.device.off(Switch.First)
        assert self.device.status().switch_1_state is False

    def test_2_off(self):
        self.device.on(Switch.Second)  # ensure on
        assert self.device.status().switch_2_state is True

        self.device.off(Switch.Second)
        assert self.device.status().switch_2_state is False

    def test_status(self):
        status = self.device.status()

        assert status.switch_1_state is _INITIAL_STATE["switch_1_state"]
        assert status.switch_1_off_delay == _INITIAL_STATE["switch_1_off_delay"]
        assert status.switch_1_default_state == _INITIAL_STATE["switch_1_default_state"]
        assert status.switch_1_state is _INITIAL_STATE["switch_1_state"]
        assert status.switch_1_off_delay == _INITIAL_STATE["switch_1_off_delay"]
        assert status.switch_1_default_state == _INITIAL_STATE["switch_1_default_state"]
        assert status.interlock == _INITIAL_STATE["interlock"]
        assert status.flex_mode == _INITIAL_STATE["flex_mode"]
        assert status.rc_list == _INITIAL_STATE["rc_list"]

    def test_set_switch_off_delay(self):
        self.device.set_switch_off_delay(300, Switch.First)
        assert self.device.status().switch_1_off_delay == 300
        self.device.set_switch_off_delay(200, Switch.Second)
        assert self.device.status().switch_2_off_delay == 200

        with pytest.raises(YeelightDualControlModuleException):
            self.device.set_switch_off_delay(-2, Switch.First)

        with pytest.raises(YeelightDualControlModuleException):
            self.device.set_switch_off_delay(43300, Switch.Second)



from unittest import TestCase

import pytest

from miio import AirConditionerMiot
from miio.airconditioner_miot import (
    AirConditionerMiotException,
    CleaningStatus,
    FanSpeed,
    OperationMode,
    TimerStatus,
)

from .dummies import DummyMiotDevice

_INITIAL_STATE = {
    "power": False,
    "mode": OperationMode.Cool,
    "target_temperature": 24,
    "eco": True,
    "heater": True,
    "dryer": False,
    "sleep_mode": False,
    "fan_speed": FanSpeed.Level7,
    "vertical_swing": True,
    "temperature": 27.5,
    "buzzer": True,
    "led": False,
    "electricity": 0.0,
    "clean": "0,100,1,1",
    "running_duration": 100.4,
    "fan_speed_percent": 90,
    "timer": "0,0,0,0",
}


class DummyAirConditionerMiot(DummyMiotDevice, AirConditionerMiot):
    def __init__(self, *args, **kwargs):
        self.state = _INITIAL_STATE
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
            "set_mode": lambda x: self._set_state("mode", x),
            "set_target_temperature": lambda x: self._set_state(
                "target_temperature", x
            ),
            "set_eco": lambda x: self._set_state("eco", x),
            "set_heater": lambda x: self._set_state("heater", x),
            "set_dryer": lambda x: self._set_state("dryer", x),
            "set_sleep_mode": lambda x: self._set_state("sleep_mode", x),
            "set_fan_speed": lambda x: self._set_state("fan_speed", x),
            "set_vertical_swing": lambda x: self._set_state("vertical_swing", x),
            "set_temperature": lambda x: self._set_state("temperature", x),
            "set_buzzer": lambda x: self._set_state("buzzer", x),
            "set_led": lambda x: self._set_state("led", x),
            "set_clean": lambda x: self._set_state("clean", x),
            "set_fan_speed_percent": lambda x: self._set_state("fan_speed_percent", x),
            "set_timer": lambda x, y: self._set_state("timer", x, y),
        }
        super().__init__(*args, **kwargs)


@pytest.fixture(scope="function")
def airconditionermiot(request):
    request.cls.device = DummyAirConditionerMiot()


@pytest.mark.usefixtures("airconditionermiot")
class TestAirConditioner(TestCase):
    def test_on(self):
        self.device.off()  # ensure off
        assert self.device.status().is_on is False

        self.device.on()
        assert self.device.status().is_on is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.device.status().is_on is True

        self.device.off()
        assert self.device.status().is_on is False

    def test_status(self):
        status = self.device.status()
        assert status.is_on == _INITIAL_STATE["power"]
        assert status.mode == OperationMode(_INITIAL_STATE["mode"])
        assert status.target_temperature == _INITIAL_STATE["target_temperature"]
        assert status.eco == _INITIAL_STATE["eco"]
        assert status.heater == _INITIAL_STATE["heater"]
        assert status.dryer == _INITIAL_STATE["dryer"]
        assert status.sleep_mode == _INITIAL_STATE["sleep_mode"]
        assert status.fan_speed == FanSpeed(_INITIAL_STATE["fan_speed"])
        assert status.vertical_swing == _INITIAL_STATE["vertical_swing"]
        assert status.temperature == _INITIAL_STATE["temperature"]
        assert status.buzzer == _INITIAL_STATE["buzzer"]
        assert status.led == _INITIAL_STATE["led"]
        assert repr(status.clean) == repr(CleaningStatus(_INITIAL_STATE["clean"]))
        assert status.fan_speed_percent == _INITIAL_STATE["fan_speed_percent"]
        assert repr(status.timer) == repr(TimerStatus(_INITIAL_STATE["timer"]))

    def test_set_mode(self):
        def mode():
            return self.device.status().mode

        self.device.set_mode(OperationMode.Cool)
        assert mode() == OperationMode.Cool

        self.device.set_mode(OperationMode.Dry)
        assert mode() == OperationMode.Dry

        self.device.set_mode(OperationMode.Fan)
        assert mode() == OperationMode.Fan

        self.device.set_mode(OperationMode.Heat)
        assert mode() == OperationMode.Heat

    def test_set_target_temperature(self):
        def target_temperature():
            return self.device.status().target_temperature

        self.device.set_target_temperature(16.0)
        assert target_temperature() == 16.0
        self.device.set_target_temperature(31.0)
        assert target_temperature() == 31.0

        with pytest.raises(AirConditionerMiotException):
            self.device.set_target_temperature(15.5)

        with pytest.raises(AirConditionerMiotException):
            self.device.set_target_temperature(24.6)

        with pytest.raises(AirConditionerMiotException):
            self.device.set_target_temperature(31.5)

    def test_set_eco(self):
        def eco():
            return self.device.status().eco

        self.device.set_eco(True)
        assert eco() is True

        self.device.set_eco(False)
        assert eco() is False

    def test_set_heater(self):
        def heater():
            return self.device.status().heater

        self.device.set_heater(True)
        assert heater() is True

        self.device.set_heater(False)
        assert heater() is False

    def test_set_dryer(self):
        def dryer():
            return self.device.status().dryer

        self.device.set_dryer(True)
        assert dryer() is True

        self.device.set_dryer(False)
        assert dryer() is False

    def test_set_sleep_mode(self):
        def sleep_mode():
            return self.device.status().sleep_mode

        self.device.set_sleep_mode(True)
        assert sleep_mode() is True

        self.device.set_sleep_mode(False)
        assert sleep_mode() is False

    def test_set_fan_speed(self):
        def fan_speed():
            return self.device.status().fan_speed

        self.device.set_fan_speed(FanSpeed.Auto)
        assert fan_speed() == FanSpeed.Auto

        self.device.set_fan_speed(FanSpeed.Level1)
        assert fan_speed() == FanSpeed.Level1

        self.device.set_fan_speed(FanSpeed.Level2)
        assert fan_speed() == FanSpeed.Level2

        self.device.set_fan_speed(FanSpeed.Level3)
        assert fan_speed() == FanSpeed.Level3

        self.device.set_fan_speed(FanSpeed.Level4)
        assert fan_speed() == FanSpeed.Level4

        self.device.set_fan_speed(FanSpeed.Level5)
        assert fan_speed() == FanSpeed.Level5

        self.device.set_fan_speed(FanSpeed.Level6)
        assert fan_speed() == FanSpeed.Level6

        self.device.set_fan_speed(FanSpeed.Level7)
        assert fan_speed() == FanSpeed.Level7

    def test_set_vertical_swing(self):
        def vertical_swing():
            return self.device.status().vertical_swing

        self.device.set_vertical_swing(True)
        assert vertical_swing() is True

        self.device.set_vertical_swing(False)
        assert vertical_swing() is False

    def test_set_led(self):
        def led():
            return self.device.status().led

        self.device.set_led(True)
        assert led() is True

        self.device.set_led(False)
        assert led() is False

    def test_set_buzzer(self):
        def buzzer():
            return self.device.status().buzzer

        self.device.set_buzzer(True)
        assert buzzer() is True

        self.device.set_buzzer(False)
        assert buzzer() is False

    def test_set_fan_speed_percent(self):
        def fan_speed_percent():
            return self.device.status().fan_speed_percent

        self.device.set_fan_speed_percent(1)
        assert fan_speed_percent() == 1
        self.device.set_fan_speed_percent(101)
        assert fan_speed_percent() == 101

        with pytest.raises(AirConditionerMiotException):
            self.device.set_fan_speed_percent(102)

        with pytest.raises(AirConditionerMiotException):
            self.device.set_fan_speed_percent(0)

    def test_set_timer(self):
        def timer():
            return self.device.status().data["timer"]

        self.device.set_timer(60, True)
        assert timer() == "1,60,1"

        self.device.set_timer(120, False)
        assert timer() == "1,120,0"

    def test_set_clean(self):
        def clean():
            return self.device.status().data["clean"]

        self.device.set_clean(True)
        assert clean() == "1"

        self.device.set_clean(False)
        assert clean() == "0"



import logging
from collections import defaultdict
from typing import Any, Dict

import click

from .click_common import command, format_output
from .device import Device, DeviceStatus
from .exceptions import DeviceException

_LOGGER = logging.getLogger(__name__)


class PhilipsEyecareException(DeviceException):
    pass


class PhilipsEyecareStatus(DeviceStatus):
    """Container for status reports from Xiaomi Philips Eyecare Smart Lamp 2."""

    def __init__(self, data: Dict[str, Any]) -> None:
        # ['power': 'off', 'bright': 5, 'notifystatus': 'off',
        #  'ambstatus': 'off', 'ambvalue': 41, 'eyecare': 'on',
        #  'scene_num': 3, 'bls': 'on', 'dvalue': 0]
        self.data = data

    @property
    def power(self) -> str:
        """Power state."""
        return self.data["power"]

    @property
    def is_on(self) -> bool:
        """True if the device is turned on."""
        return self.power == "on"

    @property
    def brightness(self) -> int:
        """Current brightness of the primary light."""
        return self.data["bright"]

    @property
    def reminder(self) -> bool:
        """Indicates the eye fatigue notification is enabled or not."""
        return self.data["notifystatus"] == "on"

    @property
    def ambient(self) -> bool:
        """True if the ambient light (second light source) is on."""
        return self.data["ambstatus"] == "on"

    @property
    def ambient_brightness(self) -> int:
        """Brightness of the ambient light."""
        return self.data["ambvalue"]

    @property
    def eyecare(self) -> bool:
        """True if the eyecare mode is on."""
        return self.data["eyecare"] == "on"

    @property
    def scene(self) -> int:
        """Current fixed scene."""
        return self.data["scene_num"]

    @property
    def smart_night_light(self) -> bool:
        """True if the smart night light mode is on."""
        return self.data["bls"] == "on"

    @property
    def delay_off_countdown(self) -> int:
        """Countdown until turning off in minutes."""
        return self.data["dvalue"]


class PhilipsEyecare(Device):
    """Main class representing Xiaomi Philips Eyecare Smart Lamp 2."""

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "Brightness: {result.brightness}\n"
            "Ambient light: {result.ambient}\n"
            "Ambient light brightness: {result.ambient_brightness}\n"
            "Eyecare mode: {result.eyecare}\n"
            "Scene: {result.scence}\n"
            "Eye fatigue reminder: {result.reminder}\n"
            "Smart night light: {result.smart_night_light}\n"
            "Delayed turn off: {result.delay_off_countdown}\n",
        )
    )
    def status(self) -> PhilipsEyecareStatus:
        """Retrieve properties."""
        properties = [
            "power",
            "bright",
            "notifystatus",
            "ambstatus",
            "ambvalue",
            "eyecare",
            "scene_num",
            "bls",
            "dvalue",
        ]
        values = self.get_properties(properties)

        return PhilipsEyecareStatus(defaultdict(lambda: None, zip(properties, values)))

    @command(default_output=format_output("Powering on"))
    def on(self):
        """Power on."""
        return self.send("set_power", ["on"])

    @command(default_output=format_output("Powering off"))
    def off(self):
        """Power off."""
        return self.send("set_power", ["off"])

    @command(default_output=format_output("Turning on eyecare mode"))
    def eyecare_on(self):
        """Turn the eyecare mode on."""
        return self.send("set_eyecare", ["on"])

    @command(default_output=format_output("Turning off eyecare mode"))
    def eyecare_off(self):
        """Turn the eyecare mode off."""
        return self.send("set_eyecare", ["off"])

    @command(
        click.argument("level", type=int),
        default_output=format_output("Setting brightness to {level}"),
    )
    def set_brightness(self, level: int):
        """Set brightness level of the primary light."""
        if level < 1 or level > 100:
            raise PhilipsEyecareException("Invalid brightness: %s" % level)

        return self.send("set_bright", [level])

    @command(
        click.argument("number", type=int),
        default_output=format_output("Setting fixed scene to {number}"),
    )
    def set_scene(self, number: int):
        """Set one of the fixed eyecare user scenes."""
        if number < 1 or number > 4:
            raise PhilipsEyecareException("Invalid fixed scene number: %s" % number)

        return self.send("set_user_scene", [number])

    @command(
        click.argument("minutes", type=int),
        default_output=format_output("Setting delayed turn off to {minutes} minutes"),
    )
    def delay_off(self, minutes: int):
        """Set delay off minutes."""

        if minutes < 0:
            raise PhilipsEyecareException(
                "Invalid value for a delayed turn off: %s" % minutes
            )

        return self.send("delay_off", [minutes])

    @command(default_output=format_output("Turning on smart night light"))
    def smart_night_light_on(self):
        """Turn the smart night light mode on."""
        return self.send("enable_bl", ["on"])

    @command(default_output=format_output("Turning off smart night light"))
    def smart_night_light_off(self):
        """Turn the smart night light mode off."""
        return self.send("enable_bl", ["off"])

    @command(default_output=format_output("Turning on eye fatigue reminder"))
    def reminder_on(self):
        """Enable the eye fatigue reminder / notification."""
        return self.send("set_notifyuser", ["on"])

    @command(default_output=format_output("Turning off eye fatigue reminder"))
    def reminder_off(self):
        """Disable the eye fatigue reminder / notification."""
        return self.send("set_notifyuser", ["off"])

    @command(default_output=format_output("Turning on ambient light"))
    def ambient_on(self):
        """Turn the ambient light on."""
        return self.send("enable_amb", ["on"])

    @command(default_output=format_output("Turning off ambient light"))
    def ambient_off(self):
        """Turn the ambient light off."""
        return self.send("enable_amb", ["off"])

    @command(
        click.argument("level", type=int),
        default_output=format_output("Setting brightness to {level}"),
    )
    def set_ambient_brightness(self, level: int):
        """Set the brightness of the ambient light."""
        if level < 1 or level > 100:
            raise PhilipsEyecareException("Invalid ambient brightness: %s" % level)

        return self.send("set_amb_bright", [level])



from unittest import TestCase

import pytest

from miio import HeaterMiot
from miio.heater_miot import HeaterMiotException, LedBrightness

from .dummies import DummyMiotDevice

_INITIAL_STATE = {
    "power": True,
    "temperature": 21.6,
    "target_temperature": 23,
    "buzzer": False,
    "led_brightness": 1,
    "child_lock": False,
    "countdown_time": 0,
}


class DummyHeaterMiot(DummyMiotDevice, HeaterMiot):
    def __init__(self, *args, **kwargs):
        self.state = _INITIAL_STATE
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
            "set_led_brightness": lambda x: self._set_state("led_brightness", x),
            "set_buzzer": lambda x: self._set_state("buzzer", x),
            "set_child_lock": lambda x: self._set_state("child_lock", x),
            "set_delay_off": lambda x: self._set_state("countdown_time", x),
            "set_target_temperature": lambda x: self._set_state(
                "target_temperature", x
            ),
        }
        super().__init__(*args, **kwargs)


@pytest.fixture(scope="class")
def heater(request):
    request.cls.device = DummyHeaterMiot()


@pytest.mark.usefixtures("heater")
class TestHeater(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def test_on(self):
        self.device.off()
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_set_led_brightness(self):
        def led_brightness():
            return self.device.status().led_brightness

        self.device.set_led_brightness(LedBrightness.On)
        assert led_brightness() == LedBrightness.On

        self.device.set_led_brightness(LedBrightness.Off)
        assert led_brightness() == LedBrightness.Off

    def test_set_buzzer(self):
        def buzzer():
            return self.device.status().buzzer

        self.device.set_buzzer(True)
        assert buzzer() is True

        self.device.set_buzzer(False)
        assert buzzer() is False

    def test_set_child_lock(self):
        def child_lock():
            return self.device.status().child_lock

        self.device.set_child_lock(True)
        assert child_lock() is True

        self.device.set_child_lock(False)
        assert child_lock() is False

    def test_set_delay_off(self):
        def delay_off_countdown():
            return self.device.status().delay_off_countdown

        self.device.set_delay_off(0)
        assert delay_off_countdown() == 0
        self.device.set_delay_off(9 * 3600)
        assert delay_off_countdown() == 9
        self.device.set_delay_off(12 * 3600)
        assert delay_off_countdown() == 12
        self.device.set_delay_off(9 * 3600 + 1)
        assert delay_off_countdown() == 9

        with pytest.raises(HeaterMiotException):
            self.device.set_delay_off(-1)

        with pytest.raises(HeaterMiotException):
            self.device.set_delay_off(13 * 3600)

    def test_set_target_temperature(self):
        def target_temperature():
            return self.device.status().target_temperature

        self.device.set_target_temperature(18)
        assert target_temperature() == 18

        self.device.set_target_temperature(23)
        assert target_temperature() == 23

        self.device.set_target_temperature(28)
        assert target_temperature() == 28

        with pytest.raises(HeaterMiotException):
            self.device.set_target_temperature(17)

        with pytest.raises(HeaterMiotException):
            self.device.set_target_temperature(29)



import json
import os
import string
from unittest import TestCase

import pytest

from miio import (
    AirConditioningCompanion,
    AirConditioningCompanionMcn02,
    AirConditioningCompanionV3,
)
from miio.airconditioningcompanion import (
    MODEL_ACPARTNER_V3,
    STORAGE_SLOT_ID,
    AirConditioningCompanionException,
    AirConditioningCompanionStatus,
    FanSpeed,
    Led,
    OperationMode,
    Power,
    SwingMode,
)
from miio.airconditioningcompanionMCN import MODEL_ACPARTNER_MCN02
from miio.airconditioningcompanionMCN import (
    AirConditioningCompanionStatus as AirConditioningCompanionStatusMcn02,
)
from miio.airconditioningcompanionMCN import FanSpeed as FanSpeedMcn02
from miio.airconditioningcompanionMCN import OperationMode as OperationModeMcn02
from miio.airconditioningcompanionMCN import SwingMode as SwingModeMcn02
from miio.tests.dummies import DummyDevice

STATE_ON = ["on"]
STATE_OFF = ["off"]

PUBLIC_ENUMS = {
    "OperationMode": OperationMode,
    "FanSpeed": FanSpeed,
    "Power": Power,
    "SwingMode": SwingMode,
    "Led": Led,
}


def as_enum(d):
    if "__enum__" in d:
        name, member = d["__enum__"].split(".")
        return getattr(PUBLIC_ENUMS[name], member)
    else:
        return d


with open(
    os.path.join(os.path.dirname(__file__), "test_airconditioningcompanion.json")
) as inp:
    test_data = json.load(inp, object_hook=as_enum)


class EnumEncoder(json.JSONEncoder):
    def default(self, obj):
        if type(obj) in PUBLIC_ENUMS.values():
            return {"__enum__": str(obj)}
        return json.JSONEncoder.default(self, obj)


class DummyAirConditioningCompanion(DummyDevice, AirConditioningCompanion):
    def __init__(self, *args, **kwargs):
        self.state = ["010500978022222102", "01020119A280222221", "2"]
        self.last_ir_played = None

        self.return_values = {
            "get_model_and_state": self._get_state,
            "start_ir_learn": lambda x: True,
            "end_ir_learn": lambda x: True,
            "get_ir_learn_result": lambda x: True,
            "send_ir_code": lambda x: self._send_input_validation(x),
            "send_cmd": lambda x: self._send_input_validation(x),
            "set_power": lambda x: self._set_power(x),
        }
        self.start_state = self.state.copy()
        super().__init__(args, kwargs)

    def _reset_state(self):
        """Revert back to the original state."""
        self.state = self.start_state.copy()

    def _get_state(self, props):
        """Return the requested data."""
        return self.state

    def _set_power(self, value: str):
        """Set the requested power state."""
        if value == STATE_ON:
            self.state[1] = self.state[1][:2] + "1" + self.state[1][3:]

        if value == STATE_OFF:
            self.state[1] = self.state[1][:2] + "0" + self.state[1][3:]

    @staticmethod
    def _hex_input_validation(payload):
        return all(c in string.hexdigits for c in payload[0])

    def _send_input_validation(self, payload):
        if self._hex_input_validation(payload[0]):
            self.last_ir_played = payload[0]
            return True

        return False

    def get_last_ir_played(self):
        return self.last_ir_played


@pytest.fixture(scope="class")
def airconditioningcompanion(request):
    request.cls.device = DummyAirConditioningCompanion()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("airconditioningcompanion")
class TestAirConditioningCompanion(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(
            AirConditioningCompanionStatus(
                dict(model_and_state=self.device.start_state)
            )
        )

        assert self.is_on() is False
        assert self.state().power_socket is None
        assert self.state().load_power == 2
        assert self.state().air_condition_model == bytes.fromhex("010500978022222102")
        assert self.state().model_format == 1
        assert self.state().device_type == 5
        assert self.state().air_condition_brand == int("0097", 16)
        assert self.state().air_condition_remote == int("80222221", 16)
        assert self.state().state_format == 2
        assert self.state().air_condition_configuration == "020119A2"
        assert self.state().target_temperature == 25
        assert self.state().swing_mode == SwingMode.Off
        assert self.state().fan_speed == FanSpeed.Low
        assert self.state().mode == OperationMode.Auto
        assert self.state().led is False

    def test_status_without_target_temperature(self):
        self.device._reset_state()
        self.device.state[1] = None
        assert self.state().target_temperature is None

    def test_status_without_swing_mode(self):
        self.device._reset_state()
        self.device.state[1] = None
        assert self.state().swing_mode is None

    def test_status_without_mode(self):
        self.device._reset_state()
        self.device.state[1] = None
        assert self.state().mode is None

    def test_status_without_fan_speed(self):
        self.device._reset_state()
        self.device.state[1] = None
        assert self.state().fan_speed is None

    def test_learn(self):
        assert self.device.learn(STORAGE_SLOT_ID) is True
        assert self.device.learn() is True

    def test_learn_result(self):
        assert self.device.learn_result() is True

    def test_learn_stop(self):
        assert self.device.learn_stop(STORAGE_SLOT_ID) is True
        assert self.device.learn_stop() is True

    def test_send_ir_code(self):
        for args in test_data["test_send_ir_code_ok"]:
            with self.subTest():
                self.device._reset_state()
                self.assertTrue(self.device.send_ir_code(*args["in"]))
                self.assertSequenceEqual(self.device.get_last_ir_played(), args["out"])

        for args in test_data["test_send_ir_code_exception"]:
            with pytest.raises(AirConditioningCompanionException):
                self.device.send_ir_code(*args["in"])

    def test_send_command(self):
        assert self.device.send_command("0000000") is True

    def test_send_configuration(self):

        for args in test_data["test_send_configuration_ok"]:
            with self.subTest():
                self.device._reset_state()
                self.assertTrue(self.device.send_configuration(*args["in"]))
                self.assertSequenceEqual(self.device.get_last_ir_played(), args["out"])


class DummyAirConditioningCompanionV3(DummyDevice, AirConditioningCompanionV3):
    def __init__(self, *args, **kwargs):
        self.state = ["010507950000257301", "011001160100002573", "807"]
        self.device_prop = {"lumi.0": {"plug_state": ["on"]}}
        self.model = MODEL_ACPARTNER_V3
        self.last_ir_played = None

        self.return_values = {
            "get_model_and_state": self._get_state,
            "get_device_prop": self._get_device_prop,
            "toggle_plug": self._toggle_plug,
        }
        self.start_state = self.state.copy()
        self.start_device_prop = self.device_prop.copy()
        super().__init__(args, kwargs)

    def _reset_state(self):
        """Revert back to the original state."""
        self.state = self.start_state.copy()

    def _get_state(self, props):
        """Return the requested data."""
        return self.state

    def _get_device_prop(self, props):
        """Return the requested data."""
        return self.device_prop[props[0]][props[1]]

    def _toggle_plug(self, props):
        """Toggle the lumi.0 plug state."""
        self.device_prop["lumi.0"]["plug_state"] = [props.pop()]


@pytest.fixture(scope="class")
def airconditioningcompanionv3(request):
    request.cls.device = DummyAirConditioningCompanionV3()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("airconditioningcompanionv3")
class TestAirConditioningCompanionV3(TestCase):
    def state(self):
        return self.device.status()

    def is_on(self):
        return self.device.status().is_on

    def test_socket_on(self):
        self.device.socket_off()  # ensure off
        assert self.state().power_socket == "off"

        self.device.socket_on()
        assert self.state().power_socket == "on"

    def test_socket_off(self):
        self.device.socket_on()  # ensure on
        assert self.state().power_socket == "on"

        self.device.socket_off()
        assert self.state().power_socket == "off"

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(
            AirConditioningCompanionStatus(
                dict(
                    model_and_state=self.device.start_state,
                    power_socket=self.device.start_device_prop["lumi.0"]["plug_state"][
                        0
                    ],
                )
            )
        )

        assert self.is_on() is True
        assert self.state().power_socket == "on"
        assert self.state().load_power == 807
        assert self.state().air_condition_model == bytes.fromhex("010507950000257301")
        assert self.state().model_format == 1
        assert self.state().device_type == 5
        assert self.state().air_condition_brand == int("0795", 16)
        assert self.state().air_condition_remote == int("00002573", 16)
        assert self.state().state_format == 1
        assert self.state().air_condition_configuration == "10011601"
        assert self.state().target_temperature == 22
        assert self.state().swing_mode == SwingMode.Off
        assert self.state().fan_speed == FanSpeed.Low
        assert self.state().mode == OperationMode.Heat
        assert self.state().led is True


class DummyAirConditioningCompanionMcn02(DummyDevice, AirConditioningCompanionMcn02):
    def __init__(self, *args, **kwargs):
        self.state = ["on", "cool", 28, "small_fan", "on", 441.0]
        self.model = MODEL_ACPARTNER_MCN02

        self.return_values = {"get_prop": self._get_state}
        self.start_state = self.state.copy()
        super().__init__(args, kwargs)

    def _reset_state(self):
        """Revert back to the original state."""
        self.state = self.start_state.copy()

    def _get_state(self, props):
        """Return the requested data."""
        return self.state


@pytest.fixture(scope="class")
def airconditioningcompanionMcn02(request):
    request.cls.device = DummyAirConditioningCompanionMcn02()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("airconditioningcompanionMcn02")
class TestAirConditioningCompanionMcn02(TestCase):
    def state(self):
        return self.device.status()

    def is_on(self):
        return self.device.status().is_on

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(
            AirConditioningCompanionStatusMcn02(self.device.start_state)
        )

        assert self.is_on() is True
        assert self.state().target_temperature == 28
        assert self.state().swing_mode == SwingModeMcn02.On
        assert self.state().fan_speed == FanSpeedMcn02.Low
        assert self.state().mode == OperationModeMcn02.Cool



import enum
from typing import Any, Dict

import click

from .click_common import EnumType, command, format_output
from .fan_common import FanException, MoveDirection, OperationMode
from .miot_device import DeviceStatus, MiotDevice

MODEL_FAN_P9 = "dmaker.fan.p9"
MODEL_FAN_P10 = "dmaker.fan.p10"
MODEL_FAN_P11 = "dmaker.fan.p11"
MODEL_FAN_1C = "dmaker.fan.1c"

MIOT_MAPPING = {
    MODEL_FAN_1C: {
        # https://miot-spec.org/miot-spec-v2/instance?type=urn:miot-spec-v2:device:fan:0000A005:dmaker-1c:1
        "power": {"siid": 2, "piid": 1},
        "fan_level": {"siid": 2, "piid": 2},
        "child_lock": {"siid": 3, "piid": 1},
        "swing_mode": {"siid": 2, "piid": 3},
        "power_off_time": {"siid": 2, "piid": 10},
        "buzzer": {"siid": 2, "piid": 11},
        "light": {"siid": 2, "piid": 12},
        "mode": {"siid": 2, "piid": 7},
    },
    MODEL_FAN_P9: {
        # Source https://miot-spec.org/miot-spec-v2/instance?type=urn:miot-spec-v2:device:fan:0000A005:dmaker-p9:1
        "power": {"siid": 2, "piid": 1},
        "fan_level": {"siid": 2, "piid": 2},
        "child_lock": {"siid": 3, "piid": 1},
        "fan_speed": {"siid": 2, "piid": 11},
        "swing_mode": {"siid": 2, "piid": 5},
        "swing_mode_angle": {"siid": 2, "piid": 6},
        "power_off_time": {"siid": 2, "piid": 8},
        "buzzer": {"siid": 2, "piid": 7},
        "light": {"siid": 2, "piid": 9},
        "mode": {"siid": 2, "piid": 4},
        "set_move": {"siid": 2, "piid": 10},
    },
    MODEL_FAN_P10: {
        # Source https://miot-spec.org/miot-spec-v2/instance?type=urn:miot-spec-v2:device:fan:0000A005:dmaker-p10:1
        "power": {"siid": 2, "piid": 1},
        "fan_level": {"siid": 2, "piid": 2},
        "child_lock": {"siid": 3, "piid": 1},
        "fan_speed": {"siid": 2, "piid": 10},
        "swing_mode": {"siid": 2, "piid": 4},
        "swing_mode_angle": {"siid": 2, "piid": 5},
        "power_off_time": {"siid": 2, "piid": 6},
        "buzzer": {"siid": 2, "piid": 8},
        "light": {"siid": 2, "piid": 7},
        "mode": {"siid": 2, "piid": 3},
        "set_move": {"siid": 2, "piid": 9},
    },
    MODEL_FAN_P11: {
        # Source https://miot-spec.org/miot-spec-v2/instance?type=urn:miot-spec-v2:device:fan:0000A005:dmaker-p11:1
        "power": {"siid": 2, "piid": 1},
        "fan_level": {"siid": 2, "piid": 2},
        "mode": {"siid": 2, "piid": 3},
        "swing_mode": {"siid": 2, "piid": 4},
        "swing_mode_angle": {"siid": 2, "piid": 5},
        "fan_speed": {"siid": 2, "piid": 6},
        "light": {"siid": 4, "piid": 1},
        "buzzer": {"siid": 5, "piid": 1},
        # "device_fault": {"siid": 6, "piid": 2},
        "child_lock": {"siid": 7, "piid": 1},
        "power_off_time": {"siid": 3, "piid": 1},
        "set_move": {"siid": 6, "piid": 1},
    },
}

SUPPORTED_ANGLES = {
    MODEL_FAN_P9: [30, 60, 90, 120, 150],
    MODEL_FAN_P10: [30, 60, 90, 120, 140],
    MODEL_FAN_P11: [30, 60, 90, 120, 140],
}


class OperationModeMiot(enum.Enum):
    Normal = 0
    Nature = 1


class FanStatusMiot(DeviceStatus):
    """Container for status reports for Xiaomi Mi Smart Pedestal Fan DMaker P9/P10."""

    def __init__(self, data: Dict[str, Any]) -> None:
        """
        Response of a FanMiot (dmaker.fan.p10):

        {
          'id': 1,
          'result': [
            {'did': 'power', 'siid': 2, 'piid': 1, 'code': 0, 'value': False},
            {'did': 'fan_level', 'siid': 2, 'piid': 2, 'code': 0, 'value': 2},
            {'did': 'child_lock', 'siid': 3, 'piid': 1, 'code': 0, 'value': False},
            {'did': 'fan_speed', 'siid': 2, 'piid': 10, 'code': 0, 'value': 54},
            {'did': 'swing_mode', 'siid': 2, 'piid': 4, 'code': 0, 'value': False},
            {'did': 'swing_mode_angle', 'siid': 2, 'piid': 5, 'code': 0, 'value': 30},
            {'did': 'power_off_time', 'siid': 2, 'piid': 6, 'code': 0, 'value': 0},
            {'did': 'buzzer', 'siid': 2, 'piid': 8, 'code': 0, 'value': False},
            {'did': 'light', 'siid': 2, 'piid': 7, 'code': 0, 'value': True},
            {'did': 'mode', 'siid': 2, 'piid': 3, 'code': 0, 'value': 0},
            {'did': 'set_move', 'siid': 2, 'piid': 9, 'code': -4003}
          ],
          'exe_time': 280
        }
        """
        self.data = data

    @property
    def power(self) -> str:
        """Power state."""
        return "on" if self.data["power"] else "off"

    @property
    def is_on(self) -> bool:
        """True if device is currently on."""
        return self.data["power"]

    @property
    def mode(self) -> OperationMode:
        """Operation mode."""
        return OperationMode[OperationModeMiot(self.data["mode"]).name]

    @property
    def speed(self) -> int:
        """Speed of the motor."""
        return self.data["fan_speed"]

    @property
    def oscillate(self) -> bool:
        """True if oscillation is enabled."""
        return self.data["swing_mode"]

    @property
    def angle(self) -> int:
        """Oscillation angle."""
        return self.data["swing_mode_angle"]

    @property
    def delay_off_countdown(self) -> int:
        """Countdown until turning off in minutes."""
        return self.data["power_off_time"]

    @property
    def led(self) -> bool:
        """True if LED is turned on, if available."""
        return self.data["light"]

    @property
    def buzzer(self) -> bool:
        """True if buzzer is turned on."""
        return self.data["buzzer"]

    @property
    def child_lock(self) -> bool:
        """True if child lock is on."""
        return self.data["child_lock"]


class FanStatus1C(DeviceStatus):
    """Container for status reports for Xiaomi Mi Smart Pedestal Fan DMaker 1C."""

    def __init__(self, data: Dict[str, Any]) -> None:
        self.data = data
        """
        Response of a Fan1C (dmaker.fan.1c):

        {
          'id': 1,
          'result': [
            {'did': 'power', 'siid': 2, 'piid': 1, 'code': 0, 'value': True},
            {'did': 'fan_level', 'siid': 2, 'piid': 2, 'code': 0, 'value': 2},
            {'did': 'child_lock', 'siid': 3, 'piid': 1, 'code': 0, 'value': False},
            {'did': 'swing_mode', 'siid': 2, 'piid': 3, 'code': 0, 'value': False},
            {'did': 'power_off_time', 'siid': 2, 'piid': 10, 'code': 0, 'value': 0},
            {'did': 'buzzer', 'siid': 2, 'piid': 11, 'code': 0, 'value': False},
            {'did': 'light', 'siid': 2, 'piid': 12, 'code': 0, 'value': True},
            {'did': 'mode', 'siid': 2, 'piid': 7, 'code': 0, 'value': 0},
          ],
          'exe_time': 280
        }
        """

    @property
    def power(self) -> str:
        """Power state."""
        return "on" if self.data["power"] else "off"

    @property
    def is_on(self) -> bool:
        """True if device is currently on."""
        return self.data["power"]

    @property
    def mode(self) -> OperationMode:
        """Operation mode."""
        return OperationMode[OperationModeMiot(self.data["mode"]).name]

    @property
    def speed(self) -> int:
        """Speed of the motor."""
        return self.data["fan_level"]

    @property
    def oscillate(self) -> bool:
        """True if oscillation is enabled."""
        return self.data["swing_mode"]

    @property
    def delay_off_countdown(self) -> int:
        """Countdown until turning off in minutes."""
        return self.data["power_off_time"]

    @property
    def led(self) -> bool:
        """True if LED is turned on."""
        return self.data["light"]

    @property
    def buzzer(self) -> bool:
        """True if buzzer is turned on."""
        return self.data["buzzer"]

    @property
    def child_lock(self) -> bool:
        """True if child lock is on."""
        return self.data["child_lock"]


class FanMiot(MiotDevice):
    mapping = MIOT_MAPPING[MODEL_FAN_P10]

    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
        model: str = MODEL_FAN_P10,
    ) -> None:
        if model not in MIOT_MAPPING:
            raise FanException("Invalid FanMiot model: %s" % model)

        super().__init__(ip, token, start_id, debug, lazy_discover)
        self.model = model

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "Operation mode: {result.mode}\n"
            "Speed: {result.speed}\n"
            "Oscillate: {result.oscillate}\n"
            "Angle: {result.angle}\n"
            "LED: {result.led}\n"
            "Buzzer: {result.buzzer}\n"
            "Child lock: {result.child_lock}\n"
            "Power-off time: {result.delay_off_countdown}\n",
        )
    )
    def status(self) -> FanStatusMiot:
        """Retrieve properties."""
        return FanStatusMiot(
            {
                prop["did"]: prop["value"] if prop["code"] == 0 else None
                for prop in self.get_properties_for_mapping()
            }
        )

    @command(default_output=format_output("Powering on"))
    def on(self):
        """Power on."""
        return self.set_property("power", True)

    @command(default_output=format_output("Powering off"))
    def off(self):
        """Power off."""
        return self.set_property("power", False)

    @command(
        click.argument("mode", type=EnumType(OperationMode)),
        default_output=format_output("Setting mode to '{mode.value}'"),
    )
    def set_mode(self, mode: OperationMode):
        """Set mode."""
        return self.set_property("mode", OperationModeMiot[mode.name].value)

    @command(
        click.argument("speed", type=int),
        default_output=format_output("Setting speed to {speed}"),
    )
    def set_speed(self, speed: int):
        """Set speed."""
        if speed < 0 or speed > 100:
            raise FanException("Invalid speed: %s" % speed)

        return self.set_property("fan_speed", speed)

    @command(
        click.argument("angle", type=int),
        default_output=format_output("Setting angle to {angle}"),
    )
    def set_angle(self, angle: int):
        """Set the oscillation angle."""
        if angle not in SUPPORTED_ANGLES[self.model]:
            raise FanException(
                "Unsupported angle. Supported values: "
                + ", ".join("{0}".format(i) for i in SUPPORTED_ANGLES[self.model])
            )

        return self.set_property("swing_mode_angle", angle)

    @command(
        click.argument("oscillate", type=bool),
        default_output=format_output(
            lambda oscillate: "Turning on oscillate"
            if oscillate
            else "Turning off oscillate"
        ),
    )
    def set_oscillate(self, oscillate: bool):
        """Set oscillate on/off."""
        if oscillate:
            return self.set_property("swing_mode", True)
        else:
            return self.set_property("swing_mode", False)

    @command(
        click.argument("led", type=bool),
        default_output=format_output(
            lambda led: "Turning on LED" if led else "Turning off LED"
        ),
    )
    def set_led(self, led: bool):
        """Turn led on/off."""
        if led:
            return self.set_property("light", True)
        else:
            return self.set_property("light", False)

    @command(
        click.argument("buzzer", type=bool),
        default_output=format_output(
            lambda buzzer: "Turning on buzzer" if buzzer else "Turning off buzzer"
        ),
    )
    def set_buzzer(self, buzzer: bool):
        """Set buzzer on/off."""
        if buzzer:
            return self.set_property("buzzer", True)
        else:
            return self.set_property("buzzer", False)

    @command(
        click.argument("lock", type=bool),
        default_output=format_output(
            lambda lock: "Turning on child lock" if lock else "Turning off child lock"
        ),
    )
    def set_child_lock(self, lock: bool):
        """Set child lock on/off."""
        return self.set_property("child_lock", lock)

    @command(
        click.argument("minutes", type=int),
        default_output=format_output("Setting delayed turn off to {minutes} minutes"),
    )
    def delay_off(self, minutes: int):
        """Set delay off minutes."""

        if minutes < 0 or minutes > 480:
            raise FanException("Invalid value for a delayed turn off: %s" % minutes)

        return self.set_property("power_off_time", minutes)

    @command(
        click.argument("direction", type=EnumType(MoveDirection)),
        default_output=format_output("Rotating the fan to the {direction}"),
    )
    def set_rotate(self, direction: MoveDirection):
        """Rotate fan to given direction."""
        # Values for: P9,P10,P11,P15,P18,...
        # { "value": 0, "description": "NONE" },
        # { "value": 1, "description": "LEFT" },
        # { "value": 2, "description": "RIGHT" }
        value = 0
        if direction == MoveDirection.Left:
            value = 1
        elif direction == MoveDirection.Right:
            value = 2
        return self.set_property("set_move", value)


class FanP9(FanMiot):
    mapping = MIOT_MAPPING[MODEL_FAN_P9]


class FanP10(FanMiot):
    mapping = MIOT_MAPPING[MODEL_FAN_P10]


class FanP11(FanMiot):
    mapping = MIOT_MAPPING[MODEL_FAN_P11]


class Fan1C(MiotDevice):
    mapping = MIOT_MAPPING[MODEL_FAN_1C]

    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
        model: str = MODEL_FAN_1C,
    ) -> None:
        super().__init__(ip, token, start_id, debug, lazy_discover)
        self.model = model

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "Operation mode: {result.mode}\n"
            "Speed: {result.speed}\n"
            "Oscillate: {result.oscillate}\n"
            "LED: {result.led}\n"
            "Buzzer: {result.buzzer}\n"
            "Child lock: {result.child_lock}\n"
            "Power-off time: {result.delay_off_countdown}\n",
        )
    )
    def status(self) -> FanStatus1C:
        """Retrieve properties."""
        return FanStatus1C(
            {
                prop["did"]: prop["value"] if prop["code"] == 0 else None
                for prop in self.get_properties_for_mapping()
            }
        )

    @command(default_output=format_output("Powering on"))
    def on(self):
        """Power on."""
        return self.set_property("power", True)

    @command(default_output=format_output("Powering off"))
    def off(self):
        """Power off."""
        return self.set_property("power", False)

    @command(
        click.argument("mode", type=EnumType(OperationMode)),
        default_output=format_output("Setting mode to '{mode.value}'"),
    )
    def set_mode(self, mode: OperationMode):
        """Set mode."""
        return self.set_property("mode", OperationModeMiot[mode.name].value)

    @command(
        click.argument("speed", type=int),
        default_output=format_output("Setting speed to {speed}"),
    )
    def set_speed(self, speed: int):
        """Set speed."""
        if speed not in (1, 2, 3):
            raise FanException("Invalid speed: %s" % speed)

        return self.set_property("fan_level", speed)

    @command(
        click.argument("oscillate", type=bool),
        default_output=format_output(
            lambda oscillate: "Turning on oscillate"
            if oscillate
            else "Turning off oscillate"
        ),
    )
    def set_oscillate(self, oscillate: bool):
        """Set oscillate on/off."""
        return self.set_property("swing_mode", oscillate)

    @command(
        click.argument("led", type=bool),
        default_output=format_output(
            lambda led: "Turning on LED" if led else "Turning off LED"
        ),
    )
    def set_led(self, led: bool):
        """Turn led on/off."""
        return self.set_property("light", led)

    @command(
        click.argument("buzzer", type=bool),
        default_output=format_output(
            lambda buzzer: "Turning on buzzer" if buzzer else "Turning off buzzer"
        ),
    )
    def set_buzzer(self, buzzer: bool):
        """Set buzzer on/off."""
        return self.set_property("buzzer", buzzer)

    @command(
        click.argument("lock", type=bool),
        default_output=format_output(
            lambda lock: "Turning on child lock" if lock else "Turning off child lock"
        ),
    )
    def set_child_lock(self, lock: bool):
        """Set child lock on/off."""
        return self.set_property("child_lock", lock)

    @command(
        click.argument("minutes", type=int),
        default_output=format_output("Setting delayed turn off to {minutes} minutes"),
    )
    def delay_off(self, minutes: int):
        """Set delay off minutes."""

        if minutes < 0 or minutes > 480:
            raise FanException("Invalid value for a delayed turn off: %s" % minutes)

        return self.set_property("power_off_time", minutes)



from unittest import TestCase

import pytest

from miio import ChuangmiPlug
from miio.chuangmi_plug import (
    MODEL_CHUANGMI_PLUG_M1,
    MODEL_CHUANGMI_PLUG_V1,
    MODEL_CHUANGMI_PLUG_V3,
    ChuangmiPlugStatus,
)

from .dummies import DummyDevice


class DummyChuangmiPlugV1(DummyDevice, ChuangmiPlug):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_CHUANGMI_PLUG_V1
        self.state = {"on": True, "usb_on": True, "temperature": 32}
        self.return_values = {
            "get_prop": self._get_state,
            "set_on": lambda x: self._set_state_basic("on", True),
            "set_off": lambda x: self._set_state_basic("on", False),
            "set_usb_on": lambda x: self._set_state_basic("usb_on", True),
            "set_usb_off": lambda x: self._set_state_basic("usb_on", False),
        }
        self.start_state = self.state.copy()
        super().__init__(args, kwargs)

    def _set_state_basic(self, var, value):
        """Set a state of a variable."""
        self.state[var] = value


@pytest.fixture(scope="class")
def chuangmiplugv1(request):
    request.cls.device = DummyChuangmiPlugV1()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("chuangmiplugv1")
class TestChuangmiPlugV1(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(ChuangmiPlugStatus(self.device.start_state))

        assert self.is_on() is True
        assert self.state().usb_power is True
        assert self.state().temperature == self.device.start_state["temperature"]

    def test_usb_on(self):
        self.device.usb_off()  # ensure off
        assert self.device.status().usb_power is False

        self.device.usb_on()
        assert self.device.status().usb_power is True

    def test_usb_off(self):
        self.device.usb_on()  # ensure on
        assert self.device.status().usb_power is True

        self.device.usb_off()
        assert self.device.status().usb_power is False


class DummyChuangmiPlugV3(DummyDevice, ChuangmiPlug):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_CHUANGMI_PLUG_V3
        self.state = {"on": True, "usb_on": True, "temperature": 32, "wifi_led": "off"}
        self.return_values = {
            "get_prop": self._get_state,
            "get_power": self._get_load_power,
            "set_power": lambda x: self._set_state_basic("on", x == ["on"]),
            "set_usb_on": lambda x: self._set_state_basic("usb_on", True),
            "set_usb_off": lambda x: self._set_state_basic("usb_on", False),
            "set_wifi_led": lambda x: self._set_state("wifi_led", x),
        }
        self.start_state = self.state.copy()
        super().__init__(args, kwargs)

    def _set_state_basic(self, var, value):
        """Set a state of a variable."""
        self.state[var] = value

    def _get_load_power(self, props=None):
        """Return load power."""
        return [300]


@pytest.fixture(scope="class")
def chuangmiplugv3(request):
    request.cls.device = DummyChuangmiPlugV3()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("chuangmiplugv3")
class TestChuangmiPlugV3(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        load_power = float(self.device._get_load_power().pop(0) * 0.01)

        start_state_extended = self.device.start_state.copy()
        start_state_extended["load_power"] = load_power
        assert repr(self.state()) == repr(ChuangmiPlugStatus(start_state_extended))

        assert self.is_on() is True
        assert self.state().usb_power is True
        assert self.state().temperature == self.device.start_state["temperature"]
        assert self.state().load_power == load_power

    def test_usb_on(self):
        self.device.usb_off()  # ensure off
        assert self.device.status().usb_power is False

        self.device.usb_on()
        assert self.device.status().usb_power is True

    def test_usb_off(self):
        self.device.usb_on()  # ensure on
        assert self.device.status().usb_power is True

        self.device.usb_off()
        assert self.device.status().usb_power is False

    def test_set_wifi_led(self):
        def wifi_led():
            return self.device.status().wifi_led

        self.device.set_wifi_led(True)
        assert wifi_led() is True

        self.device.set_wifi_led(False)
        assert wifi_led() is False


class DummyChuangmiPlugM1(DummyDevice, ChuangmiPlug):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_CHUANGMI_PLUG_M1
        self.state = {"power": "on", "temperature": 32}
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
        }
        super().__init__(args, kwargs)


@pytest.fixture(scope="class")
def chuangmiplugm1(request):
    request.cls.device = DummyChuangmiPlugM1()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("chuangmiplugm1")
class TestChuangmiPlugM1(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(ChuangmiPlugStatus(self.device.start_state))

        assert self.is_on() is True
        assert self.state().temperature == self.device.start_state["temperature"]



import pytest
from miio.deviceinfo import DeviceInfo


@pytest.fixture()
def info():
    """Example response from Xiaomi Smart WiFi Plug (c&p from deviceinfo ctor)."""
    return DeviceInfo(
        {
            "ap": {"bssid": "FF:FF:FF:FF:FF:FF", "rssi": -68, "ssid": "network"},
            "cfg_time": 0,
            "fw_ver": "1.2.4_16",
            "hw_ver": "MW300",
            "life": 24,
            "mac": "28:FF:FF:FF:FF:FF",
            "mmfree": 30312,
            "model": "chuangmi.plug.m1",
            "netif": {
                "gw": "192.168.xxx.x",
                "localIp": "192.168.xxx.x",
                "mask": "255.255.255.0",
            },
            "ot": "otu",
            "ott_stat": [0, 0, 0, 0],
            "otu_stat": [320, 267, 3, 0, 3, 742],
            "token": "2b00042f7481c7b056c4b410d28f33cf",
            "wifi_fw_ver": "SD878x-14.76.36.p84-702.1.0-WM",
        }
    )


def test_properties(info):
    """Test that all deviceinfo properties are accessible."""

    assert info.raw == info.data

    assert isinstance(info.accesspoint, dict)
    assert isinstance(info.network_interface, dict)

    ap_props = ["bssid", "ssid", "rssi"]
    for prop in ap_props:
        assert prop in info.accesspoint

    if_props = ["gw", "localIp", "mask"]
    for prop in if_props:
        assert prop in info.network_interface

    assert info.model is not None
    assert info.firmware_version is not None
    assert info.hardware_version is not None
    assert info.mac_address is not None


def test_missing_fields(info):
    """Test that missing keys do not cause exceptions."""
    for k in ["fw_ver", "hw_ver", "model", "token", "mac"]:
        del info.raw[k]

    assert info.model is None
    assert info.firmware_version is None
    assert info.hardware_version is None
    assert info.mac_address is None
    assert info.token is None



import logging
from enum import IntEnum

import click

from .click_common import command, format_output
from .device import Device

_LOGGER = logging.getLogger(__name__)

MODEL = "scishare.coffee.s1102"


class Status(IntEnum):
    Unknown = -1
    Off = 1
    On = 2
    SelfCheck = 3
    StopPreheat = 4
    CoffeeReady = 5
    StopDescaling = 6
    Standby = 7
    Preheating = 8

    Brewing = 201
    NoWater = 203


class ScishareCoffee(Device):
    """Main class for Scishare coffee maker (scishare.coffee.s1102)."""

    @command()
    def status(self) -> int:
        """Device status."""
        status_code = self.send("Query_Machine_Status")[1]
        try:
            return Status(status_code)
        except ValueError:
            _LOGGER.warning(
                "Status code unknown, please report the state of the machine for code %s",
                status_code,
            )
            return Status.Unknown

    @command(
        click.argument("temperature", type=int),
        default_output=format_output("Setting preheat to {temperature}"),
    )
    def preheat(self, temperature: int):
        """Pre-heat to given temperature."""
        return self.send("Boiler_Preheating_Set", [temperature])

    @command(default_output=format_output("Stopping pre-heating"))
    def stop_preheat(self) -> bool:
        """Stop pre-heating."""
        return self.send("Stop_Boiler_Preheat")[0] == "ok"

    @command()
    def cancel_alarm(self) -> bool:
        """Unknown."""
        raise NotImplementedError()
        return self.send("Cancel_Work_Alarm")[0] == "ok"

    @command(
        click.argument("amount", type=int),
        click.argument("temperature", type=int),
        default_output=format_output("Boiling {amount} ml water ({temperature}C)"),
    )
    def boil_water(self, amount: int, temperature: int) -> bool:
        """Boil water.

        :param amount: in milliliters
        :param temperature: in degrees
        """
        return self.send("Hot_Wate", [amount, temperature])[0] == "ok"

    @command(
        click.argument("amount", type=int),
        click.argument("temperature", type=int),
        default_output=format_output("Brewing {amount} ml espresso ({temperature}C)"),
    )
    def brew_espresso(self, amount: int, temperature: int):
        """Brew espresso.

        :param amount: in milliliters
        :param temperature: in degrees
        """
        return self.send("Espresso_Coffee", [amount, temperature])[0] == "ok"

    @command(
        click.argument("water_amount", type=int),
        click.argument("water_temperature", type=int),
        click.argument("coffee_amount", type=int),
        click.argument("coffee_temperature", type=int),
        default_output=format_output(
            "Brewing americano using {water_amount} ({water_temperature}C) water and {coffee_amount} ml ({coffee_temperature}C) coffee"
        ),
    )
    def brew_americano(
        self,
        water_amount: int,
        water_temperature: int,
        coffee_amount: int,
        coffee_temperature: int,
    ) -> bool:
        """Brew americano.

        :param water_amount: water in milliliters
        :param water_temperature: water temperature
        :param coffee_amount: coffee amount in milliliters
        :param coffee_temperature: coffee temperature
        """
        return (
            self.send(
                "Americano_Coffee",
                [water_amount, water_temperature, coffee_amount, coffee_temperature],
            )[0]
            == "ok"
        )

    @command(default_output=format_output("Powering on"))
    def on(self) -> bool:
        """Power on."""
        return self.send("Machine_ON")[0] == "ok"

    @command(default_output=format_output("Powering off"))
    def off(self) -> bool:
        """Power off."""
        return self.send("Machine_OFF")[0] == "ok"

    @command()
    def buzzer_frequency(self):
        """Unknown."""
        raise NotImplementedError()
        return self.send("Buzzer_Frequency_Time")[0] == "ok"



from unittest import TestCase

import pytest

from miio import DreameVacuumMiot
from miio.dreamevacuum_miot import (
    ChargingState,
    CleaningMode,
    DeviceStatus,
    FaultStatus,
    OperatingMode,
)

from .dummies import DummyMiotDevice

_INITIAL_STATE = {
    "battery_level": 42,
    "charging_state": ChargingState.Charging,
    "device_fault": FaultStatus.NoFaults,
    "device_status": DeviceStatus.Paused,
    "brush_left_time": 235,
    "brush_life_level": 85,
    "filter_life_level": 66,
    "filter_left_time": 154,
    "brush_left_time2": 187,
    "brush_life_level2": 57,
    "operating_mode": OperatingMode.Cleaning,
    "cleaning_mode": CleaningMode.Medium,
    "delete_timer": 12,
    "life_sieve": "9000-9000",
    "life_brush_side": "12000-12000",
    "life_brush_main": "18000-18000",
    "timer_enable": "false",
    "start_time": "22:00",
    "stop_time": "8:00",
    "deg": 5,
    "speed": 5,
    "map_view": "tmp",
    "frame_info": 3,
    "volume": 4,
    "voice_package": "DE",
}


class DummyDreameVacuumMiot(DummyMiotDevice, DreameVacuumMiot):
    def __init__(self, *args, **kwargs):
        self.state = _INITIAL_STATE
        super().__init__(*args, **kwargs)


@pytest.fixture(scope="function")
def dummydreamevacuum(request):
    request.cls.device = DummyDreameVacuumMiot()


@pytest.mark.usefixtures("dummydreamevacuum")
class TestDreameVacuum(TestCase):
    def test_status(self):
        status = self.device.status()
        assert status.battery_level == _INITIAL_STATE["battery_level"]
        assert status.brush_left_time == _INITIAL_STATE["brush_left_time"]
        assert status.brush_left_time2 == _INITIAL_STATE["brush_left_time2"]
        assert status.brush_life_level2 == _INITIAL_STATE["brush_life_level2"]
        assert status.brush_life_level == _INITIAL_STATE["brush_life_level"]
        assert status.filter_left_time == _INITIAL_STATE["filter_left_time"]
        assert status.filter_life_level == _INITIAL_STATE["filter_life_level"]
        assert status.device_fault == FaultStatus(_INITIAL_STATE["device_fault"])
        assert repr(status.device_fault) == repr(
            FaultStatus(_INITIAL_STATE["device_fault"])
        )
        assert status.charging_state == ChargingState(_INITIAL_STATE["charging_state"])
        assert repr(status.charging_state) == repr(
            ChargingState(_INITIAL_STATE["charging_state"])
        )
        assert status.operating_mode == OperatingMode(_INITIAL_STATE["operating_mode"])
        assert repr(status.operating_mode) == repr(
            OperatingMode(_INITIAL_STATE["operating_mode"])
        )
        assert status.cleaning_mode == CleaningMode(_INITIAL_STATE["cleaning_mode"])
        assert repr(status.cleaning_mode) == repr(
            CleaningMode(_INITIAL_STATE["cleaning_mode"])
        )
        assert status.device_status == DeviceStatus(_INITIAL_STATE["device_status"])
        assert repr(status.device_status) == repr(
            DeviceStatus(_INITIAL_STATE["device_status"])
        )
        assert status.life_sieve == _INITIAL_STATE["life_sieve"]
        assert status.life_brush_side == _INITIAL_STATE["life_brush_side"]
        assert status.life_brush_main == _INITIAL_STATE["life_brush_main"]
        assert status.timer_enable == _INITIAL_STATE["timer_enable"]
        assert status.start_time == _INITIAL_STATE["start_time"]
        assert status.stop_time == _INITIAL_STATE["stop_time"]
        assert status.map_view == _INITIAL_STATE["map_view"]
        assert status.volume == _INITIAL_STATE["volume"]
        assert status.voice_package == _INITIAL_STATE["voice_package"]
